#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
GzipPackager.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Gzip a file.

"""

from ebookmaker.packagers import OneFileGzipPackager

TYPE = 'gzip'
FORMATS = 'rst html.noimages html.images txt.us-ascii txt.iso-8859-1 txt.utf-8'.split ()

class Packager (OneFileGzipPackager):
    """ Gzip packager. """
    pass



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

RSTParser.py

Copyright 2010-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""

from __future__ import unicode_literals

# FIXME:
# use docinfo instead of meta for pg header

import re
from functools import partial

from six.moves import urllib

from lxml import etree
import lxml.html

import docutils.readers.standalone
from docutils import nodes, frontend
import docutils

from pkg_resources import resource_string # pylint: disable=E0611

from libgutenberg.GutenbergGlobals import NS
from libgutenberg.Logger import info, debug, warning, error
from libgutenberg.MediaTypes import mediatypes as mt

from ebookmaker import ParserFactory, parsers
from ebookmaker.parsers import HTMLParser, BROKEN

from ebookmaker.mydocutils import nodes as mynodes
from ebookmaker.mydocutils.writers import xhtml1, epub2, xetex

from ebookmaker.mydocutils.gutenberg import parsers as gutenberg_parsers
from ebookmaker.mydocutils.gutenberg.writers import nroff as gutenberg_nroff

mediatypes = (mt.rst, )

REB_EMACS_CHARSET = re.compile (rb'-\*-.*coding:\s*(\S+)',  re.I)

class Parser (HTMLParser.Parser):
    """ Parse a ReStructured Text

    and convert it to different xhtml flavours.

    """

    def __init__ (self, attribs = None):
        HTMLParser.Parser.__init__ (self, attribs)
        self.document1 = None


    def preprocess (self, charset):
        """ Insert pg header and footer. """

        return self.unicode_content ()


    def to_xhtml (self, html, base_url):
        html = html.replace ('&nbsp;', ' ')
        html = html.replace ('&mdash;', '—')

        try:
            xhtml = etree.fromstring (
                html,
                lxml.html.XHTMLParser (),
                base_url = base_url)
        except etree.ParseError as what:
            error ("etree.fromstring says %s" % what)
            raise

        xhtml.make_links_absolute (base_url = base_url)

        return xhtml


    def rewrite_links (self, f):
        """ Rewrite all links using the function f. """

        doc = self.document1

        if 'coverpage' in doc.meta_block:
            coverpage = doc.meta_block['coverpage']
            coverpage[0] = f (coverpage[0])
        else:
            for field in doc.traverse (nodes.field):
                field_name, field_body = field.children
                if field_name.astext () == 'coverpage':
                    field_body[:] = nodes.paragraph ('', f (field_body.astext ()))
                    break

        for node in doc.traverse (nodes.reference):
            if 'uri' in node:
                node['uri'] = f (node['uri'])

        for node in doc.traverse (nodes.image):
            if 'uri' in node:
                node['uri'] = f (node['uri'])

        for node in doc.traverse (nodes.pending):
            # dropcap images
            if 'image' in node.details:
                node.details['image'] = f (node.details['image'])


    def iterlinks (self):
        """ Grab links and images in RST. """

        debug ("RSTParser iterlinks")

        doc = self.document1

        # return coverpage even in no_images build
        if 'coverpage' in doc.meta_block:
            coverpage = doc.meta_block['coverpage']
            yield coverpage[0], parsers.em.link (rel='icon')
        else:
            for field in doc.traverse (nodes.field):
                field_name, field_body = field.children
                if field_name.astext () == 'coverpage':
                    yield field_body.astext(), parsers.em.link (rel='icon')
                    break

        # need broken.png for no_images build
        # yield (BROKEN), parsers.em.img (rel = 'important')

        # anchor links
        for node in doc.traverse (nodes.reference):
            if 'uri' in node:
                yield node['uri'], parsers.em.a ()

        # images
        for node in doc.traverse (nodes.image):
            d = parsers.em.img ()
            # specify class important if you want to keep an image in a no_images build
            if 'important' in node.get ('classes', []):
                d.set('rel', 'important')
            if 'uri' in node:
                yield node['uri'], d

        # dropcap images
        for node in doc.traverse (nodes.pending):
            if 'image' in node.details:
                yield node.details['image'], parsers.em.img ()


    def get_settings (self, components, defaults):
        option_parser = frontend.OptionParser (
            components = components,
            defaults = defaults,
            read_config_files = 1)
        return option_parser.get_default_values ()


    def pre_parse (self):
        """ Parse a RST file as link list. """

        debug ("RSTParser: Pre-parsing %s" % self.attribs.url)

        default_style = self.get_resource (
            'mydocutils.parsers', 'default_style.rst')

        source = docutils.io.StringInput (default_style + self.unicode_content ())
        reader = docutils.readers.standalone.Reader ()
        parser = gutenberg_parsers.Parser ()

        overrides = {
            'get_resource': self.get_resource,
            'get_image_size': self.get_image_size_from_parser,
            'base_url': self.attribs.url,
            }

        doc = reader.read (
            source, parser, self.get_settings ((reader, parser), overrides))
        self.document1 = doc

        self.rewrite_links (partial (urllib.parse.urljoin, self.attribs.url))

        debug ("RSTParser: Done pre-parsing %s" % self.attribs.url)


    def _full_parse (self, writer, overrides):
        """ Full parse from scratch. """

        debug ("RSTParser: Full-parsing %s" % self.attribs.url)

        default_style = self.get_resource (
            'mydocutils.parsers', 'default_style.rst')

        source = docutils.io.StringInput (default_style + self.unicode_content (),
                                 self.attribs.url, 'unicode')
        reader = docutils.readers.standalone.Reader ()
        parser = gutenberg_parsers.Parser ()

        doc = reader.read (
            source, parser,
            self.get_settings ((reader, parser, writer), overrides))
        self.document1 = doc

        self.rewrite_links (partial (urllib.parse.urljoin, self.attribs.url))

        doc.transformer.populate_from_components ((source, reader, parser, writer))
        doc.transformer.apply_transforms ()
        debug ("RSTParser: Done full-parsing %s" % self.attribs.url)

        return doc


    def _full_parse_2 (self, writer, destination, overrides):
        """ Full parser from pickled doctree.

        Doesn't work yet. It turned out pickling a doctree is much
        harder than I thought. """

        debug ("Full-parsing %s" % self.attribs.url)

        source = docutils.io.StringInput (self.unicode_content ())
        reader = docutils.readers.standalone.Reader ()
        parser = gutenberg_parsers.Parser ()

        doc = reader.read (
            source, parser,
            self.get_settings ((reader, parser, writer), overrides))
        self.document1 = doc

        self.rewrite_links (partial (urllib.parse.urljoin, self.attribs.url))

        # make it picklable
        reporter = doc.reporter #  = None
        # doc.reporter = None
        transformer = doc.transformer
        doc.settings = None
        from docutils.parsers.rst.directives.html import MetaBody

        #for metanode in doc.traverse (MetaBody.meta):
        for pending in doc.traverse (nodes.pending):
            # pending.transform = None
            # docutils' meta nodes aren't picklable because the class is nested
            # in pending['nodes']
            if 'nodes' in pending.details:
                if isinstance (pending.details['nodes'][0], MetaBody.meta):
                    pending.details['nodes'][0].__class__ = mynodes.meta
        from six.moves import cPickle as pickle
        pickled = pickle.dumps (doc)

        doc = pickle.loads (pickled)

        #doc.transformer.populate_from_components (
        #    (source, reader, parser, writer))

        doc.transformer = transformer
        doc.reporter = reporter
        doc.settings = self.get_settings ((reader, parser, writer), overrides)

        doc.transformer.apply_transforms ()

        return writer.write (doc, destination)


    def rst2nroff (self, job, charset = 'utf-8'):
        """ Convert RST to nroff. """

        writer = gutenberg_nroff.Writer ()
        destination = docutils.io.StringOutput (encoding = 'unicode') # !

        overrides = {
            'doctitle_xform': 1,
            'sectsubtitle_xform': 1,
            'footnote_references': 'superscript',
            'compact_lists': 1,
            'compact_simple': 1,
            'page_numbers': 1,
            'no_images': True,
            'get_resource': self.get_resource,
            'format': job.type,
            'encoding': charset,
            'base_url': self.attribs.url,
            }

        doc = self._full_parse (writer, overrides)
        return writer.write (doc, destination)


    def rst2xetex (self, job):
        """ Convert RST to xetex. """

        writer = xetex.Writer ()
        destination = docutils.io.StringOutput (encoding = 'utf-8')

        overrides = {
            'doctitle_xform': 1,
            'sectsubtitle_xform': 1,
            'footnote_references': 'superscript',
            'compact_lists': 1,
            'compact_simple': 1,
            'page_numbers': 1,
            'format': job.type,
            'encoding': 'utf-8',
            'get_resource': self.get_resource,
            'get_image_size': self.get_image_size_from_parser,
            'base_url': self.attribs.url,
            }

        doc = self._full_parse (writer, overrides)
        return writer.write (doc, destination)


    def rst2htmlish (self, job, writer, more_overrides = {}):

        destination = docutils.io.StringOutput (encoding = 'unicode')

        overrides = {
            'stylesheet': None,
            'stylesheet_path': None,
            'xml_declaration': 0,
            'doctitle_xform': 1,
            'initial_header_level': 2,
            'sectsubtitle_xform': 1,
            'footnote_references': 'superscript',
            'page_numbers': 1,
            'format': job.type,
            'encoding': 'utf-8',
            'get_resource': self.get_resource,
            'get_image_size': self.get_image_size_from_parser,
            'base_url': self.attribs.url,
            }
        overrides.update (more_overrides)

        doc = self._full_parse (writer, overrides)
        return writer.fixup_xhtml (self.to_xhtml (writer.write (doc, destination), self.attribs.url))


    def rst2html (self, job):
        """ Convert RST input to HTML output. """
        return self.rst2htmlish (job, xhtml1.Writer ())


    def rst2epub2 (self, job):
        """ Convert RST input to HTML output with Epub2 tweaks. """
        return self.rst2htmlish (job, epub2.Writer (),
                                 { 'toc_backlinks': 'none' })


    def get_resource (self, package, resource):
        return resource_string ('ebookmaker.' + package, resource).decode ('utf-8')


    def get_image_size_from_parser (self, uri):
        # debug ("Getting image dimen for %s" % uri)
        parser = ParserFactory.ParserFactory.create (uri)
        parser.pre_parse ()
        if hasattr (parser, 'get_image_dimen'):
            return parser.get_image_dimen ()
        return None


    def get_charset_from_meta (self):
        """ Parse text for hints about charset. """
        # .. -*- coding: utf-8 -*-

        charset = None
        rst = self.bytes_content ()

        match = REB_EMACS_CHARSET.search (rst)
        if match:
            charset = match.group (1).decode ('ascii')
            debug ('Got charset %s from emacs comment' % charset)

        return charset


    def parse (self):
        """ Dummy. Use rst2* instead. """

        debug ("Done parsing %s" % self.attribs.url)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
PDFPackager.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Package a PDF file for PG.

"""

from ebookmaker.packagers import OneFileZipPackager

TYPE = 'ww'
FORMATS = ''.split ()

class Packager (OneFileZipPackager):
    """ WW packager for PDF files. """
    pass



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

CSSParser.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Open an url and return raw data.

"""

import logging
import re
from six.moves import urllib

import cssutils

from libgutenberg.Logger import debug
from libgutenberg.MediaTypes import mediatypes as mt

from ebookmaker import parsers
from ebookmaker.parsers import ParserBase

RE_ELEMENT = re.compile(r'((?:^|\s)[a-z0-9]+)', re.I)

mediatypes = (mt.css, )

PG_CSS_PROFILE = (
    'Added Properties for Project Gutenberg',
    {
        'display': 'flex',
        'justify-content': 'center',
        'speak': r'auto|never|always',
        'speak-as': 'normal|spell-out|digits|literal-punctuation|no-punctuation',
    },
    None
)

cssutils.profile.addProfiles([PG_CSS_PROFILE])

class Parser(ParserBase):
    """ Parse an external CSS file. """

    def __init__(self, attribs=None):
        cssutils.log.setLog(logging.getLogger('cssutils'))
        # logging.DEBUG is way too verbose
        cssutils.log.setLevel(max(cssutils.log.getEffectiveLevel(), logging.INFO))
        ParserBase.__init__(self, attribs)
        self.sheet = None

    def pre_parse(self):
        """ Parse the CSS file. """

        if self.sheet is not None:
            return

        parser = cssutils.CSSParser()
        if self.fp:
            self.sheet = parser.parseString(self.unicode_content())
        else:
            try:
                self.sheet = parser.parseUrl(self.attribs.url)
            except ValueError:
                logging.error('Missing file: %s', self.attribs.url)
                return

        self.attribs.mediatype = parsers.ParserAttributes.HeaderElement('text/css')
        self.lowercase_selectors(self.sheet)
        self.make_links_absolute()


    def parse_string(self, s):
        """ Parse the CSS in string. """

        if self.sheet is not None:
            return

        parser = cssutils.CSSParser()
        self.sheet = parser.parseString(s)

        self.attribs.mediatype = parsers.ParserAttributes.HeaderElement('text/css')
        self.lowercase_selectors(self.sheet)


    @staticmethod
    def iter_properties(sheet):
        """ Iterate on properties in css. """
        for rule in sheet:
            if rule.type == rule.STYLE_RULE:
                for prop in rule.style:
                    yield prop


    @staticmethod
    def lowercase_selectors(sheet):
        """ make selectors lowercase to match xhtml tags """
        for rule in sheet:
            if rule.type == rule.STYLE_RULE:
                for sel in rule.selectorList:
                    sel.selectorText = RE_ELEMENT.sub(lambda m: m.group(1).lower(),
                                                      sel.selectorText)

    def make_links_absolute(self):
        """ make links absolute """
        def abs_url(url):
            return urllib.parse.urljoin(self.attribs.url, url)
        cssutils.replaceUrls(self.sheet, abs_url)


    def rewrite_links(self, f):
        """ Rewrite all links using the function f. """
        cssutils.replaceUrls(self.sheet, f)


    def iterlinks(self):
        """ Return the urls of all images in document."""

        for url in cssutils.getUrls(self.sheet):
            yield urllib.parse.urljoin(self.attribs.url, url), parsers.em.style()

    def strip_images(self):
        """ remove all rules with url() in them """
        to_delete = []
        for rule in self.sheet:
            if rule.type == rule.STYLE_RULE and rule.cssText and 'url(' in rule.cssText:
                to_delete.append(rule)
        for rule in to_delete:
            self.sheet.deleteRule(rule)


    def get_aux_urls(self):
        """ Return the urls of all auxiliary files in document.

        Auxiliary files are non-document files you need to correctly
        display the document file, eg. CSS files.

        """

        aux = []

        for rule in self.sheet:
            if rule.type == rule.IMPORT_RULE:
                aux.append(urllib.parse.urljoin(self.attribs.url, rule.href))

        return  aux


    def serialize(self):
        """ Serialize CSS. """

        return self.sheet.cssText



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

RST parser module.

Copyright 2010-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

A slightly modified RST parser for docutils.

"""

from __future__ import unicode_literals

import collections
import re
import itertools
from functools import partial

import six

from docutils import nodes, statemachine
from docutils.utils import unescape
import docutils.parsers.rst
from docutils.parsers.rst import Directive, directives, states, roles, frontend
from docutils.parsers.rst.directives import body
from docutils.parsers.rst.directives import tables
from docutils.parsers.rst.directives import images
from docutils.utils import punctuation_chars

from libgutenberg.Logger import error, warning, info, debug

from ebookmaker.parsers import BROKEN
from ebookmaker.mydocutils import nodes as mynodes
from ebookmaker.mydocutils.transforms import parts

# pylint: disable=W0142, W0102

image_align_values  = ('left', 'center', 'right')
image_float_values  = ('none', 'here', 'top', 'bottom', 'page')
table_hrules_values = ('none', 'table', 'rows')
table_vrules_values = ('none', 'table', 'columns')
entry_align_values  = ('left', 'center', 'right', 'justify')
entry_valign_values = ('top',  'middle', 'bottom')

def float_arg (argument):
    """
    Converts the argument into a float.  Raises ValueError for
    non-numeric values.  (Directive option conversion function.)
    """
    try:
        return float (argument)
    except ValueError:
        raise ValueError ('non-numeric value; must be numeric')

def quoted_string (argument):
    """
    Unquotes the argument into a string.  Raises ValueError for
    unquoted values.  (Directive option conversion function.)
    """

    m = re.match ("^'(.*?)'$", argument)
    if m:
        return m.group (1)
    raise ValueError ('non-numeric value; must be numeric')

def choice (values, argument):
    """
    Checks an actual value against a list of approved values.
    Raises ValueError if the value is not found in the list of values.
    Usage: option_spec = {'option' : functools.partial (choice, list_of_values) }
    """
    return directives.choice (argument, values)


def multi_choice (values, argument):
    """
    Checks a space-separated list of actual values against a list of approved values.
    Raises ValueError if it finds any value that is not in the list.
    Usage: option_spec = {'option' : functools.partial (multi_choice, list_of_values) }
    """
    entries = argument.split ()
    for e in entries:
        directives.choice (e, values)
    return entries


def class_option_add_remove (argument):
    """
    Convert the argument into a list of ID-compatible strings and return it.
    (Directive option conversion function.)

    Raise ``ValueError`` if no argument is found.
    """
    if argument is None:
        raise ValueError('argument required but none supplied')
    names = argument.split()
    class_names = []
    for name in names:
        if name.startswith ('-'):
            class_name = '-' + nodes.make_id(name[1:])
        else:
            class_name = nodes.make_id(name)
        if not class_name:
            raise ValueError('cannot make "%s" into a class name' % name)
        class_names.append(class_name)
    return class_names


def string2lines (astring, tab_width=8, convert_whitespace=0,
                  whitespace=re.compile('[\v\f]')):
    """
    Swiped from statemachine.py and tweaked to only rstrip U+0020 spaces.

    Return a list of one-line strings with tabs expanded, no newlines, and
    trailing whitespace stripped.

    Each tab is expanded with between 1 and `tab_width` spaces, so that the
    next character's index becomes a multiple of `tab_width` (8 by default).

    Parameters:

    - `astring`: a multi-line string.
    - `tab_width`: the number of columns between tab stops.
    - `convert_whitespace`: convert form feeds and vertical tabs to spaces?
    """
    if convert_whitespace:
        astring = whitespace.sub (' ', astring)
    return [s.expandtabs (tab_width).rstrip (' ') for s in astring.splitlines ()]


class Style (Directive):
    """ Set presentational preferences for docutils element.

    """

    optional_arguments = 1
    has_content = True

    option_spec = {
        'class': class_option_add_remove,
        'name': directives.unchanged,
        'language': directives.class_option, # FIXME: a lang not a class
        'before': quoted_string,
        'after': quoted_string,
        'display': directives.unchanged,
        'formats': directives.unchanged,
        'titlehack': directives.flag, # see StyleTransform

        # images
        # note: figwidth and figclass can be set using width and class on the figure node
        'align': partial (choice, entry_align_values + entry_valign_values),
        'width': directives.length_or_percentage_or_unitless,
        'float': partial (multi_choice, image_float_values),

        # tables
        'hrules': partial (multi_choice, table_hrules_values),
        'vrules': partial (multi_choice, table_vrules_values),
        'aligns': partial (multi_choice, entry_align_values),
        'vertical-aligns': partial (multi_choice, entry_valign_values),
        'tabularcolumns': directives.unchanged,
        'widths': directives.positive_int_list,
    }

    def run (self):
        if self.arguments:
            self.options['selector'] = self.arguments[0]
        if 'language' in self.options:
            self.options['class'].extend (['language-' + x for x in self.options['language']])
            del self.options['language']

        pending = nodes.pending (parts.StyleTransform, self.options)
        self.state_machine.document.note_pending (pending)

        if self.content:
            # parse content into pending node
            self.state.nested_parse (self.content, self.content_offset, pending)

        return [pending]


class VSpace (Directive):
    """
    Directive to insert vertical spacing or pagebreaks.
    """

    required_arguments = 0
    optional_arguments = 1

    def run (self):
        arg = self.arguments[0] if self.arguments else self.name

        if arg in ('clearpage', 'cleardoublepage', 'vfill',
                   'frontmatter', 'mainmatter', 'backmatter'):
            page = mynodes.page ()
            page['classes'].append (arg)
            return [page]
        else:
            try:
                arg = abs (int (arg))
            except ValueError:
                raise self.error ('Unknown argument "%s" for "%s" directive.' % (arg, self.name))

            if arg == 0:
                nl = mynodes.newline ()
                nl['classes'].append ('white-space-pre-line')
                return [nl]

            page = mynodes.page ()
            page['classes'].append ('vspace')
            page['length'] = arg
            return [page]


class EndSection (Directive):
    """ Closes section. """

    def run (self):
        debug ('Endsection directive state: %s' % self.state)
        # back out of lists, etc.
        if isinstance (self.state, states.SpecializedBody):
            debug ('Backing out of list')
            self.state_machine.previous_line (2) # why do we need 2 ???
        raise EOFError


class DropCap (Directive):
    """ Puts a dropcap onto the next paragraph.

    """

    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True

    option_spec = { 'class': directives.class_option,
                    'name': directives.unchanged,
                    'image': directives.unchanged,
                    'lines': directives.positive_int,
                    'raise': float_arg,
                    'indents': directives.unchanged, # array of tex dimen
                    'ante': directives.unchanged }

    def run (self):
        self.options['char'] = self.arguments[0]
        if len (self.arguments) >= 2:
            self.options['span'] = self.arguments[1]
        pending = nodes.pending (parts.DropCapTransform, self.options)
        self.state_machine.document.note_pending (pending)
        return [pending]


class Example (Directive):
    """
    Builds a literal block with the example source followed by a rendered block.
    """

    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True
    example_count = 0

    option_spec = { 'class': directives.class_option,
                    'name': directives.unchanged,
                    'float': partial (multi_choice, image_float_values),
                    'norender': directives.flag }

    def run(self):
        nodelist = []

        if len (self.arguments) >= 1:
            nodelist.append (nodes.title ('', self.arguments[0]))

        literal = nodes.literal_block ('', '\n'.join (self.content))
        literal['classes'].append ('example-source')
        nodelist.append (literal)

        if 'norender' not in self.options:
            container = nodes.container ()
            container['classes'].append ('example-rendered')
            self.state.nested_parse (self.content, self.content_offset,
                                     container)
            nodelist.append (container)

        topic = nodes.topic ('', *nodelist)
        topic['classes'] += ['example']
        topic['classes'] += self.options.get ('class', [])
        Example.example_count += 1
        target = lox_target (topic, 'example-%d' % Example.example_count)
        self.state_machine.document.note_implicit_target (target, target)
        return [topic]


def lox_target (x, id_):
    id_ = nodes.make_id (id_)
    x['ids'].append (id_)
    return x


class Table (tables.RSTTable):

    option_spec = tables.RSTTable.option_spec.copy ()
    option_spec.update ( {
        'align': partial (choice, image_align_values),
        'float': partial (multi_choice, image_float_values),
        'hrules': partial (multi_choice, table_hrules_values),
        'summary': directives.unchanged,
        'tabularcolumns': directives.unchanged,
        'vrules': partial (multi_choice, table_vrules_values),
        'width': directives.length_or_percentage_or_unitless,

        'aligns': partial (multi_choice, entry_align_values),
        'vertical-aligns': partial (multi_choice, entry_valign_values),
        'widths': directives.positive_int_list,
        } )

    table_count = 0

    def run (self):

        def apply_options_to_colspecs (option, name):
            options = self.options.get (option, [])
            for colspec, option in zip (table.traverse (nodes.colspec), options):
                colspec[name] = option

        res = tables.RSTTable.run (self)
        table = res[0]

        for name in ('align', 'float', 'hrules', 'summary', 'tabularcolumns', 'vrules', 'width'):
            if name in self.options:
                table[name] = self.options[name]

        Table.table_count += 1
        target = lox_target (table, 'table-%d' % Table.table_count)
        self.state_machine.document.note_implicit_target (target, target)

        # HTML doesn't recognize align in colspec ???
        apply_options_to_colspecs ('aligns', 'align')
        apply_options_to_colspecs ('vertical-aligns', 'valign')
        apply_options_to_colspecs ('widths', 'colwidth')

        return res


class Figure (images.Figure):
    """ Override standard figure. Allow use of no filename. """

    option_spec = images.Figure.option_spec.copy ()
    option_spec.update ( {
        'float': partial (multi_choice, image_float_values),
        } )

    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    figure_count = 0

    def run (self):
        if not self.arguments:
            # no filename given: use `broken´ picture
            self.arguments.append (BROKEN)
            self.options.setdefault ('figwidth', '80%')
            self.options.setdefault ('width',    '5%')

        res = images.Figure.run (self)

        figure = res[0]
        Figure.figure_count += 1
        target = lox_target (figure, 'figure-%d' % Figure.figure_count)
        self.state_machine.document.note_implicit_target (target, target)

        return res


class TocEntry (Directive):
    """ Directive to inject TOC entry.

    This directive changes the next header, so we can choose the
    text of the TOC entry.

    The 'depth' option changes toc gathering level.

    We insert a pending node that gets transformed in later stages of processing.

    """

    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True

    option_spec = { 'class': directives.class_option,
                    'name': directives.unchanged,
                    'depth': directives.nonnegative_int }

    def run (self):
        details = {}
        if 'depth' in self.options:
            details['toc_depth'] = self.options['depth']

        container = nodes.container ()

        if self.arguments:
            container += nodes.inline ('', self.arguments[0])

        self.state.nested_parse (self.content, self.content_offset,
                                 container)

        container['classes'] += self.options.get ('class', [])
        if container.astext ():
            details['content'] = container
        else:
            details['content'] = None # flag: supress toc entry

        pending = nodes.pending (parts.TocEntryTransform, details)
        self.state_machine.document.note_pending (pending)
        return [pending]


def option_check_backlinks (arg): # pylint: disable=E0213
    """ Argument checker function. """
    value = directives.choice (arg, ('top', 'entry', 'none'))
    return None if value == 'none' else value


class GeneratedSection (Directive):
    """ Base class for generated sections (Contents, Footnotes, Citations, etc.). """

    required_arguments = 0
    optional_arguments = 1   # optional section name
    final_argument_whitespace = True

    option_spec = { 'class': directives.class_option,
                    'name': directives.unchanged,
                    'local': directives.flag,
                    'backlinks': option_check_backlinks,
                    'page-numbers': directives.flag }

    def run (self, transform, extraclasses = None):
        """
        extra_classes: classes set on container element
        """

        document = self.state_machine.document
        extraclasses = extraclasses or []

        self.options['directive_name'] = self.name
        extraclasses.append (self.name) # contents, loa, lof, lot, footnotes

        container = nodes.container ()
        container['classes'] += self.options.get ('class', [])
        if 'depth' in self.options:
            container['toc_depth'] = self.options['depth']

        pending = nodes.pending (transform, self.options)
        document.note_pending (pending)
        container += pending

        pending = nodes.pending (parts.EmptySectionRemover)
        document.note_pending (pending)

        entries = []

        if 'local' in self.options:
            container['classes'] += ['local']
            container['classes'] += ['local-' + x for x in extraclasses]
            entries = [container]
        elif self.arguments:
            container['classes'] += extraclasses

            section = nodes.section ()
            title_text = self.arguments[0]
            text_nodes, dummy_messages = self.state.inline_text (title_text,
                                                          self.lineno)
            title = nodes.title (title_text, '', *text_nodes)
            title['toc_entry'] = None # default for generated sections is: no toc entry
            section += title
            section += pending
            section += container
            document.note_implicit_target (section)
            entries = [section]
        else:
            container['classes'] += extraclasses
            entries = [pending, container]

        return entries


class Contents (GeneratedSection):
    """ Table of contents. """

    option_spec = GeneratedSection.option_spec.copy ()
    option_spec.update ( {'depth': directives.nonnegative_int } )

    def run (self):
        return GeneratedSection.run (self, parts.ContentsTransform)


class ListOfAnything (GeneratedSection):
    """ List of Figures or List of Tables or ... """

    option_spec = GeneratedSection.option_spec.copy ()
    option_spec.update ( {'selector': directives.unchanged } )

    def run (self):
        return GeneratedSection.run (self, parts.ListOfAnythingTransform, ['loa'])


class Footnotes (GeneratedSection):
    """ Footnote section. """

    option_spec = GeneratedSection.option_spec.copy ()
    option_spec.update ( {'backlinks': directives.flag } )

    def run (self):
        return GeneratedSection.run (self, parts.FootnotesDirectiveTransform)


class Container (body.Container):
    """ Backward compatibility: reinstate Container :class: option. """

    option_spec = body.Container.option_spec.copy ()
    option_spec.update ( {'class': directives.class_option,
                          'align': partial (choice, image_align_values) } )

    def run (self):
        nodes =  body.Container.run (self)
        nodes[0]['classes'].extend (self.options.get ('class', []))
        if 'align' in self.options:
            nodes[0]['align'] = self.options['align']
        return nodes


class Meta (Directive):
    """ Collect meta fields into document.meta_block, which is a dict of lists. """

    has_content = True

    def run (self):
        self.assert_has_content ()
        doc = self.state_machine.document

        node = nodes.Element () # temp container node

        new_line_offset, blank_finish = self.state.nested_list_parse (
            self.content, self.content_offset, node,
            initial_state='FieldList', blank_finish=1)

        if (new_line_offset - self.content_offset) != len(self.content):
            # incomplete parse of block?
            error = self.state_machine.reporter.error (
                'Invalid meta directive.',
                nodes.literal_block (self.block_text, self.block_text),
                line = self.lineno)
            return [error]

        # walk children of node and extract data
        for field in node:
            if isinstance (field, nodes.field):
                field_name, field_body = field.children
                doc.meta_block[field_name.astext ()].append (field_body.astext ())
            elif isinstance (field, nodes.Bibliographic):
                doc.meta_block[field.__class__.__name__].append (field.astext ())

        return []


# roles

def id_generator (i = 0):
    """ Generate an id for the TOC to link to. """
    while True:
        yield 'autoid-%05d' % i
        i += 1

id_generator = id_generator ()

def mk_pageno_target (title, id_, class_, document):
    """ Create and initialize a target node. """

    id_ = nodes.make_id (id_)

    if id_ in document.ids:
        # already taken, make a new one
        id_ = six.next (id_generator)

    target = nodes.target ('', '')
    target['ids'].append (id_)
    target['classes'].append (class_)
    # if 'classes' in options:
    #     target['classes'].extend (options['classes'])
    target[class_] = title
    target['html_attributes'] = {'title' : title}
    target['refid'] = None # avoid transform moving this target into next section

    document.note_implicit_target (target) # pylint: disable=E1101

    return target


class Role:
    """
    Base class for our roles.
    """

    def __init__ (self, options = {}, content = []):
        self.options = { 'class': directives.class_option }
        self.options.update (options)

    def __call__ (self, role, rawtext, text, lineno, inliner, options={}, content=[]):
        """ return [nodelist], [messagelist] """
        roles.set_classes (options)
        inline = nodes.inline (rawtext, unescape (text), **options)
        inline['classes'].append (role)
        return [ inline ], []


class Inliner (states.Inliner):
    """ Inliner that recognizes [pg n], [ln n], and [var x]. """

    def __init__ (self):
        self.start_string_prefix = (u'(^|(?<=\\s|[%s%s]))' %
                               (punctuation_chars.openers,
                                punctuation_chars.delimiters))
        self.end_string_suffix = (u'($|(?=\\s|[\x00%s%s%s]))' %
                             (punctuation_chars.closing_delimiters,
                              punctuation_chars.delimiters,
                              punctuation_chars.closers))
        self.re_pageno = re.compile (self.start_string_prefix +
                                r"(\[pg!?\s*(?P<pageno>[-ivxlc\d]+)(?:\s+(?P<imgoff>[\d]+))?\])" +
                                self.end_string_suffix + r'\s*', re.IGNORECASE)
        self.re_pageref = re.compile (self.start_string_prefix +
                                r"(\[pg\s*(?P<pageno>[-ivxlc\d]+)\]_)" +
                                self.end_string_suffix, re.IGNORECASE)
        self.re_lineno = re.compile (self.start_string_prefix +
                                r"(\[ln!?\s*(?P<lineno>[\d]+)\])" +
                                self.end_string_suffix + r'\s*', re.IGNORECASE)
        self.re_lineref = re.compile (self.start_string_prefix +
                                r"(\[ln\s*(?P<lineno>[\d]+)\]_)" +
                                self.end_string_suffix, re.IGNORECASE)
        self.re_variable = re.compile (self.start_string_prefix +
                                r"(\[var\s+(?P<name>[-_\w\d]+)\])" +
                                self.end_string_suffix + r'\s*', re.IGNORECASE)
        states.Inliner.__init__ (self)


    def init_customizations (self, settings):
        """ Setting-based customizations; run when parsing begins. """
        super().init_customizations (settings)

        # self.implicit_dispatch.append ((self.re_newline, self.newline))
        self.implicit_dispatch.append ((self.re_variable, self.variable))
        if settings.page_numbers:
            # pylint: disable=E1101
            self.implicit_dispatch.append ((self.re_pageno,  self.pageno_target))
            self.implicit_dispatch.append ((self.re_pageref, self.pageno_reference))
            self.implicit_dispatch.append ((self.re_lineno,  self.lineno_target))
            self.implicit_dispatch.append ((self.re_lineref, self.lineno_reference))


    def variable (self, match, dummy_lineno):
        """ Makes `[var x]` into implicit markup for a variable named x. """
        n = mynodes.variable ()
        n['name'] = match.group ('name')
        return [n]


    def newline (self, match, dummy_lineno):
        """ Makes `[br]` into implicit markup for new line. """
        n = mynodes.newline ()
        n['classes'].append ('white-space-pre-line')
        return [n]


    def pageno_target (self, match, dummy_lineno):
        """ Makes [pg 42 69] into implicit markup for page numbers.

        42 is the page number
        69 is the index of the page into the array of page images

        """

        text = match.group (0)
        if text.startswith ('[pg'):
            title = match.group ('pageno')
            target = mk_pageno_target (title, 'page-%s' % title, 'pageno', self.document)
            imgoff = match.group ('imgoff')
            if imgoff:
                target['imgoff'] = imgoff
            if '!' in text:
                target['classes'].append ('invisible')
            return [target]
        else:
            raise states.MarkupMismatch


    def pageno_reference (self, match, dummy_lineno):
        """ Makes [pg 99]_ into implicit markup for page reference. """

        text = match.group (0)
        if text.startswith ('[pg'):
            text = match.group ('pageno')
            id_ = nodes.make_id ('page-%s' % text)
            reference = nodes.reference ('', unescape (text))
            reference['refid'] = id_
            return [reference]
        else:
            raise states.MarkupMismatch


    def lineno_target (self, match, dummy_lineno):
        """ Makes [ln 999] into implicit markup for line numbers. """

        text = match.group (0)
        if text.startswith ('[ln'):
            title = match.group ('lineno')
            target = mk_pageno_target (title, 'line-%s' % title, 'lineno', self.document)
            if '!' in text:
                target['classes'].append ('invisible')
            return [target]
        else:
            raise states.MarkupMismatch


    def lineno_reference (self, match, dummy_lineno):
        """ Makes [ln 999]_ into implicit markup for line reference. """

        text = match.group (0)
        if text.startswith ('[ln'):
            text = match.group ('lineno')
            id_ = nodes.make_id ('line-%s' % text)
            reference = nodes.reference ('', unescape (text))
            reference['refid'] = id_
            return [reference]
        else:
            raise states.MarkupMismatch

# copy states.Inliner properties to Inliner so that init_customizations method works
# this is needed because init_customizations depends on locals(), which seems like a bad practice
#
for prop, val in states.Inliner.__dict__.items():
    if not prop.startswith('_'):
        setattr(Inliner, prop, val)


class Body (states.Body):
    """ Body class with fix for line block indentation. """

    def nest_line_block_level (self, block, indent_level, indent_levels):
        """ Recursive part of nest_line_block_segment. """

        last_indent = indent_levels[indent_level]
        line_block = nodes.line_block ()
        while len (block):
            indent = block[0].indent
            if indent < last_indent:
                break # end recursion
            if indent > last_indent:
                line_block.append (
                    self.nest_line_block_level (block, indent_level + 1, indent_levels))
                continue
            line_block.append (block.pop (0))
        return line_block


    def nest_line_block_segment (self, block):
        """
        Replacement for docutils.states.Body.nest_line_block_segment
        with fix for correct indentation.
        """

        # indent_levels maps indent level to no. of leading spaces
        indent_levels = sorted (set ( [item.indent for item in block] ))

        new_block = self.nest_line_block_level (block, 0, indent_levels)
        block.replace_self (new_block)


class Parser (docutils.parsers.rst.Parser):
    """ Slightly modified rst parser. """

    def __init__ (self):

        self.settings_spec = self.settings_spec + (
            'reStructuredText Parser Options',
            None,
            (('Recognize page numbers and page references (like "[pg n] and [pg n]_").',
             ['--page-numbers'],
             {'action': 'store_true', 'validator': frontend.validate_boolean}),))

        settings_defaults = {
            'no_figures': False,
            'get_resource': None,
            'get_image_size': None,
            }

        directives.register_directive ('contents',        Contents)  # replaces standard directive
        directives.register_directive ('container',       Container) # replaces standard directive
        directives.register_directive ('lof',             ListOfAnything)
        directives.register_directive ('lot',             ListOfAnything)
        directives.register_directive ('table',           Table)     # replaces standard directive
        directives.register_directive ('figure',          Figure)    # replaces standard directive
        directives.register_directive ('meta',            Meta)      # replaces standard directive
        directives.register_directive ('toc-entry',       TocEntry)
        directives.register_directive ('footnotes',       Footnotes)
        directives.register_directive ('dropcap',         DropCap)
        directives.register_directive ('example',         Example)
        directives.register_directive ('style',           Style)

        directives.register_directive ('endsection',      EndSection)

        directives.register_directive ('vspace',          VSpace)
        directives.register_directive ('vfill',           VSpace)
        directives.register_directive ('clearpage',       VSpace)
        directives.register_directive ('cleardoublepage', VSpace)
        directives.register_directive ('frontmatter',     VSpace)
        directives.register_directive ('mainmatter',      VSpace)
        directives.register_directive ('backmatter',      VSpace)

        docutils.parsers.rst.Parser.__init__ (self)
        self.inliner = Inliner ()


    def register_local_role (self, name, role):
        """ Helper """

        r = role ()
        r.role_name = name
        roles.register_local_role (name, r)


    def parse (self, inputstring, document):
        """Parse `inputstring` and populate `document`, a document tree."""
        document.meta_block = collections.defaultdict (list)

        # following code swiped from docutils/parsers/rst/__init__.py
        # and tweaked to call our own string2lines (that only strips
        # U+0020 trailing whitespace).
        # docutils.parsers.rst.Parser.parse (self, inputstring,
        # document)

        self.setup_parse (inputstring, document)
        self.statemachine = states.RSTStateMachine (
              state_classes = self.state_classes,
              initial_state = self.initial_state,
              debug = document.reporter.debug_flag)
        inputlines = string2lines (
              inputstring, tab_width = document.settings.tab_width,
              convert_whitespace = 1)
        self.statemachine.run (inputlines, document, inliner = self.inliner)
        self.finish_parse ()


    def get_transforms (self):
        tfs = docutils.parsers.rst.Parser.get_transforms (self)
        tfs.extend ([
                parts.PageNumberMoverTransform,
                parts.TocPageNumberTransform,
                parts.TextTransform,
                parts.NodeTypeTransform,
                parts.InlineImageTransform,
                parts.SetDefaults,
                parts.AlignTransform,
                parts.BlockImageWrapper,
                parts.TitleLevelTransform,
                parts.DocInfoCollector,
                parts.FirstParagraphTransform,
                parts.Lineblock2VSpace
                ])
        return tfs


# horrible hack, substitute our bug-fixed Body class
states.state_classes = (Body, ) + states.state_classes[1:]



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

ParserFactory.py

Copyright 2009-14 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""


import os.path
import re

from six.moves import urllib
import six

from pkg_resources import resource_listdir, resource_stream # pylint: disable=E0611
import requests

from libgutenberg.Logger import critical, debug, error, info
from libgutenberg import MediaTypes
from ebookmaker.CommonCode import Options
from ebookmaker.Version import VERSION
from ebookmaker import parsers

options = Options()
parser_modules = {}

def load_parsers():
    """ See what types we can parse. """

    for fn in resource_listdir('ebookmaker.parsers', ''):
        modulename, ext = os.path.splitext(fn)
        if ext == '.py':
            if modulename.endswith('Parser'):
                module = __import__('ebookmaker.parsers.' + modulename, fromlist=[modulename])
                debug("Loading parser from module: %s for mediatypes: %s" % (
                    modulename, ', '.join(module.mediatypes)))
                for mediatype in module.mediatypes:
                    parser_modules[mediatype] = module

    return parser_modules.keys()


def unload_parsers():
    """ Unload parser modules. """
    for k in parser_modules.keys():
        del parser_modules[k]


class ParserFactory(object):
    """ A factory and a cache for parsers.

    So we don't reparse the same file twice.

    """

    parsers = {} # cache: parsers[url] = parser

    @staticmethod
    def get(attribs):
        """ Get the right kind of parser. """

        try:
            mediatype = attribs.orig_mediatype.value
            if mediatype == 'text/plain' and attribs.referrer:
                # don't use GutenbergTextParser, it's a linked text file
                return parsers.TxtParser(attribs)
            return parser_modules[mediatype].Parser(attribs)
        except (AttributeError, KeyError):
            return parser_modules['*/*'].Parser(attribs)


    @classmethod
    def create(cls, url, attribs=None):
        """ Create an appropriate parser. """
        url = parsers.webify_url(url)
        if attribs is None:
            attribs = parsers.ParserAttributes()

        # debug("Need parser for %s" % url)

        if url in cls.parsers:
            # debug("... reusing parser for %s" % url)
            # reuse same parser, maybe already filled with data
            parser = cls.parsers[url]
            parser.reset()
            parser.attribs.update(attribs)
            # debug(str(parser.attribs))
            return parser

        scheme = urllib.parse.urlsplit(url).scheme
        if scheme == 'resource':
            fp = cls.open_resource(url, attribs)
        elif scheme in ('http', 'https'):
            fp = cls.open_url(url, attribs)
        else:
            fp = cls.open_file(url, attribs)
        if fp is None:
            return
        if attribs.url in cls.parsers:
            # reuse parser because parsing may be expensive, eg. reST docs
            # debug("... reusing parser for %s" % attribs.url)
            parser = cls.parsers[attribs.url]
            parser.attribs.update(attribs)
            return parser

        # ok. so we have to create a new parser
        debug("... creating new parser for %s" % url)

        if options.mediatype_from_extension:
            attribs.orig_mediatype = attribs.HeaderElement(MediaTypes.guess_type(url))
            debug("... set mediatype %s from extension" % attribs.orig_mediatype.value)

        attribs.orig_url = url
        parser = cls.get(attribs)
        parser.fp = fp

        cls.parsers[url] = parser

        return parser


    @classmethod
    def open_url(cls, url, attribs):
        """ Open url for parsing. """

        fp = requests.get(
            url,
            stream=True,
            headers={
                'User-Agent': "EbookMaker/%s (+http://pypi.python.org/ebookmaker)" % VERSION
            },
            proxies=options.config.PROXIES
        )
        attribs.orig_mediatype = attribs.HeaderElement.from_str(
            fp.headers.get('Content-Type', 'application/octet-stream'))
        debug("... got mediatype %s from server" % str(attribs.orig_mediatype))
        attribs.orig_url = url
        attribs.url = fp.url
        return six.BytesIO(fp.content)


    @classmethod
    def open_file(cls, url, attribs):
        """ Open a local file for parsing. """
        def open_file_from_path(path):
            try:
                return open(url, 'rb')
            except FileNotFoundError:
                critical('Missing file: %s' % url)
            except IsADirectoryError:
                critical('Missing file is a directory: %s' % url)
            return None
            
        if re.search(r'^([a-zA-z]:|/)', url):
            fp = open_file_from_path(url)
        else:
            try:
                # handles all the flavors of file: urls, including on windows
                fp = urllib.request.urlopen(url)
            except urllib.error.URLError as what:
                fp = None
                critical('Missing file: %s' % what.reason)
                return None
            except ValueError:  # just a relative path?
                fp = open_file_from_path(url)
            
        attribs.orig_mediatype = attribs.HeaderElement(MediaTypes.guess_type(url))

        debug("... got mediatype %s from guess_type" % str(attribs.orig_mediatype))
        attribs.orig_url = attribs.url = url
        return fp


    @classmethod
    def open_resource(cls, orig_url, attribs):
        """ Open a python package resource file for parsing. """

        # resource://python.package/filename.ext

        o = urllib.parse.urlsplit(orig_url)
        package = o.hostname
        filename = o.path[1:]
        fp = resource_stream(package, filename)
        attribs.orig_mediatype = attribs.HeaderElement(MediaTypes.guess_type(filename))

        debug("... got mediatype %s from guess_type" % str(attribs.orig_mediatype))
        attribs.orig_url = orig_url
        attribs.url = orig_url
        return fp


    @classmethod
    def clear_parser_cache(cls):
        """ Clear parser cache to free memory. """

        # debug: kill refs
        for dummy_url, parser in cls.parsers.items():
            del parser

        cls.parsers = {}



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

Epub3Writer.py

Copyright 2022 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

Writes an EPUB3 file.

"""

import re
import datetime
import os
import copy
from xml.sax.saxutils import quoteattr

from six.moves import urllib
from lxml import etree
from lxml.builder import ElementMaker

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import NS, xpath
from libgutenberg.Logger import critical, debug, error, exception, info, warning
from libgutenberg.MediaTypes import mediatypes as mt

from ebookmaker import parsers
from ebookmaker import ParserFactory
from ebookmaker import HTMLChunker
from ebookmaker.CommonCode import Options
from ebookmaker.Version import VERSION, GENERATOR

from .EpubWriter import (
    MAX_IMAGE_SIZE,
    MAX_COVER_DIMEN,
    MAX_IMAGE_DIMEN,
    LINKED_IMAGE_SIZE,
    LINKED_IMAGE_DIMEN,
    OPS_CONTENT_DOCUMENTS,
    OPS_FONT_TYPES,
    OutlineFixer,
    EPUB_TYPE,
    STRIP_CLASSES,
    TocNCX
)
HANDHELD_QUERY = 'max-width: 480px'
from . import EpubWriter, HTMLWriter

options = Options()


match_link_url = re.compile(r'^https?://', re.I)
match_non_link = re.compile(r'[a-zA-Z0-9_\-\.]*(#.*)?$')


PRIVATE_CSS = """\
@charset "utf-8";

body, body.tei.tei-text {
   color: black;
   background-color: white;
   width: auto;
   border: 0;
   padding: 0;
   }
div, p, pre, h1, h2, h3, h4, h5, h6 {
   margin-left: 0;
   margin-right: 0;
   display: block
   }
section.pgheader{
    page-break-after: always;
    }
section.pgfooter{
    page-break-before: always;
    }
div.pgebub-root-div {
   margin: 0
   }
h2 {
   page-break-before: always;
   padding-top: 1em
   }
div.figcenter span.caption {
   display: block;
   }
.pgmonospaced {
   font-family: monospace;
   font-size: 0.9em
   }
a.pgkilled {
   text-decoration: none;
   }
img.x-ebookmaker-cover {max-width: 100%;}

@media (orientation: landscape) {
    img.x-ebookmaker-cover {height: 100%;}
    }
@media (orientation: portrait) {
    img.x-ebookmaker-cover {width: 100%;}
    }
"""

class OEBPSContainer(EpubWriter.OEBPSContainer):
    """ Class representing an OEBPS Container. """


    def add_image_wrapper(self, img_url, img_title):
        """ Add a HTML file wrapping img_url. """
        img_title = quoteattr(img_title)
        filename = 'wrap%04d.xhtml' % self.wrappers
        self.wrappers += 1
        self.add_bytes(filename,
                       parsers.IMAGE_WRAPPER.format(src=img_url,
                                                    title=img_title,
                                                    backlink="",
                                                    wrapper_class='x-ebookmaker-cover',
                                                    doctype=gg.HTML5_DOCTYPE,
                                                    style=parsers.STYLE_LINK),
                       mt.xhtml)
        return filename


class Toc(TocNCX):
    """ Class that builds toc.xhtml. derived from EpubWriter.TocNCX"""

    def __init__(self, dc):
        self.toc = []
        self.dc = dc
        self.seen_urls = {}
        self.elementmaker = ElementMaker(namespace=str(NS.xhtml),
                                         nsmap={None: str(NS.xhtml), 'epub': str(NS.epub)})


    def __unicode__(self):
        """ Serialize toc.em as unicode string. """
        em = self.elementmaker
        tocdepth = 1

        if self.toc:
            self.normalize_toc()

            tocdepth = max(t[2] for t in self.toc)

        doc_title = em.title(self.dc.title)
        head = em.head(
            doc_title,
            em.meta(name='dtb:uid', content=self.dc.opf_identifier),
            em.meta(name='dtb:depth', content=str(tocdepth)),
            em.meta(name='dtb:generator', content=GENERATOR % VERSION),
            em.meta(name='dtb:totalPageCount', content='0'),
            em.meta(name='dtb:maxPageNumber', content='0'))

        body = em.body(**{EPUB_TYPE: 'frontmatter'})

        self.seen_urls = {}
        has_pages = False
        for url, dummy_title, depth in self.toc:
            # navPoints and pageTargets referencing the same element
            # must have the same playOrder
            if url not in self.seen_urls:
                self.seen_urls[url] = str(len(self.seen_urls) + 1)
            if depth == -1:
                has_pages = True

        params = {NS.xml.lang: self.dc.languages[0].id} if self.dc.languages else {}

        body.append(self._make_navmap(self.toc))
        ncx = em.html(
            head,
            body,
            **params
        )

        if has_pages:
            body.append(self._make_pagelist(self.toc))

        # Ugly workaround for error: "Serialisation to unicode must not
        # request an XML declaration"

        toc_ncx = "%s\n\n%s" % (gg.XML_DECLARATION, etree.tostring(ncx,
                                                                   doctype=None,
                                                                   encoding=str,
                                                                   pretty_print=True))

        if options.verbose >= 3:
            debug(toc_ncx)
        return toc_ncx


    def _make_navmap(self, toc):
        """ Build the toc. """
        em = self.elementmaker

        root = em.nav(**{EPUB_TYPE: 'toc'})
        toctop = em.ol()
        root.append(toctop)

        count = 0
        prev_depth = 1
        current_ol = toctop
        for url, title, depth in toc:
            if depth < 0:
                continue
            count += 1
            toc_item = em.a(title, **{'href': url, 'id': "np-%d" % count})
            if depth > prev_depth:
                while depth > prev_depth:
                    for el in reversed(current_ol):
                        li = el
                        break
                    else:
                        li = em.li()
                    new_ol = em.ol()
                    li.append(new_ol)
                    prev_depth += 1
                    current_ol.append(li)
                    current_ol = new_ol
            else:
                while depth < prev_depth:
                    current_ol = current_ol.getparent().getparent()
                    prev_depth -= 1
            li = em.li(toc_item)
            current_ol.append(li)

        return root


    def _make_pagelist(self, toc):
        """ Build the page list. """
        em = self.elementmaker
        root = em.nav(**{EPUB_TYPE: 'landmarks'})
        pagelist_top = em.ol(**{'id': 'pages', 'class': 'pagelist'})
        root.append(pagelist_top)

        for url, pagename, depth in toc:
            if depth == -1:
                toc_item = em.a(pagename, **{
                    'href': url,
                    'id': "pt-%d" % len(pagelist_top),
                    EPUB_TYPE: 'normal' if re.search('[0-9]', pagename) else 'frontmatter',
                })
                pagelist_top.append(em.li(toc_item))

        return root


class ContentOPF(object):
    """ Class that builds content.opf metadata. """

    def __init__(self):
        self.nsmap = gg.build_nsmap('opf dc dcterms xsi')

        # FIXME: remove this when lxml is fixed
        # workaround for lxml fat-fingering the default attribute namespaces
        self.nsmap[None] = str(NS.opf) + 'lxml-bug-workaround'

        self.opf = ElementMaker(namespace=self.nsmap[None], nsmap=self.nsmap)

        self.metadata = self.opf.metadata()
        self.manifest = self.opf.manifest()
        self.spine = self.opf.spine()
        self.item_id = 0


    def __str__(self):
        return self.__unicode__()


    def __unicode__(self):
        """ Serialize content.opf as unicode string. """

        assert len(self.manifest) > 0, 'No manifest item in content.opf.'
        assert len(self.spine) > 0, 'No spine item in content.opf.'

        package = self.opf.package(
            **{'version': '3.0', 'unique-identifier': 'id'}) # FIXME add version to instance
        package.append(self.metadata)
        package.append(self.manifest)
        package.append(self.spine)

        content_opf = "%s\n\n%s" % (gg.XML_DECLARATION,
                                    etree.tostring(package,
                                                   encoding=str,
                                                   pretty_print=True))

        # FIXME: remove this when lxml is fixed
        # now merge xmlns:opf and xmlns:
        content_opf = content_opf.replace('lxml-bug-workaround', '')

        if options.verbose >= 3:
            debug(content_opf)
        return content_opf


    def rewrite_links(self, f):
        """ Rewrite all links through f(). """
        for item in self.manifest:
            if item.get('href'):
                item.set('href', f(item.get('href')))


    def meta_item(self, name, content):
        """ Add item to metadata. """
        self.metadata.append(self.opf.meta(name=name, content=content))


    def manifest_item(self, url, mediatype, id_=None, prop=None):
        """ Add item to manifest. """
        def add_prop(prop, newprop):
            if prop:
                vals = prop.split()
            else:
                vals = []
            vals.append(newprop)
            prop = ' '.join(vals)
            return prop

        if id_ is None or xpath(self.manifest, "//*[@id = '%s']" % id_):
            self.item_id += 1
            id_ = 'item%d' % self.item_id

        if prop == 'cover':
            self.add_coverpage(url, id_)
        manifest_atts = {'href': url, 'id': id_, 'media-type': mediatype}
        if mediatype == 'image/svg+xml':
            prop = add_prop(prop, 'svg')
        if prop:
            manifest_atts['properties'] = prop
        self.manifest.append(
            self.opf.item(**manifest_atts))

        return id_


    def spine_item(self, url, mediatype, id_=None, first=False):
        """ Add item to spine and manifest. """

        if id_ and id_.startswith('pgepubid'):
            # this is an auto-generated header id, not human-readable and probably duplicated
            # make a new one
            id_ = None

        id_ = self.manifest_item(url, mediatype, id_)

        # HACK: ADE needs cover flow as first element
        # but we don't know if we have a native coverpage until the manifest is complete
        if first:
            self.spine.insert(
                0, self.opf.itemref(idref=id_))
        else:
            self.spine.append(
                self.opf.itemref(idref=id_))


    def manifest_item_from_parser(self, p):
        """ Add item to manifest from parser. """
        if hasattr(p.attribs, 'comment'):
            self.manifest.append(etree.Comment(p.attribs.comment))
        cover = 'cover-image' if 'icon' in p.attribs.rel else None
        return self.manifest_item(p.attribs.url, p.mediatype(), id_=p.attribs.id, prop=cover)


    def spine_item_from_parser(self, p):
        """ Add item to spine and manifest from parser. """
        if hasattr(p.attribs, 'comment'):
            self.manifest.append(etree.Comment(p.attribs.comment))
        return self.spine_item(p.attribs.url, p.mediatype(), p.attribs.id)


    def toc_item(self, url):
        """ Add TOC to manifest and spine. """
        self.manifest_item(url, mt.xhtml, id_='ncx', prop='nav')


    def pagemap_item(self, url):
        """ Add page-map to manifest and spine. """
        self.manifest_item(url, 'application/oebps-page-map+xml', 'map')
        self.spine.attrib['page-map'] = 'map'


    def metadata_item(self, dc):
        """ Build metadata from DublinCore struct.

        Example of metadata:

  <metadata xmlns:dcterms='http://purl.org/dc/terms/'
            xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'
            xmlns:opf='http://www.idpf.org/2007/opf'>

    <dc:identifier id='id'>http://www.gutenberg.org/ebooks/29000</dc:identifier>
    <dc:creator id="author_0">Anthony Trollope</dc:creator>
    <meta property="file-as" refines="#author_0">Trollope, Anthony</meta>
    <meta property="role" refines="#author_0" scheme="marc:relators">aut</meta>
    <dc:title>The Macdermots of Ballycloran</dc:title>
    <dc:language>en</dc:language>
    <dc:subject>Domestic fiction</dc:subject>
    <dc:subject>Ireland -- Fiction</dc:subject>
    <dc:created>1890</dc:created>
    <dc:publisher>Project Gutenberg</dc:publisher>
    <dc:date opf:event='publication'>2009-05-31</dc:date>
    <dc:date opf:event='conversion'>2009-08-26T21:11:14Z</dc:date>
    <dc:rights>Public domain</dc:rights>
    <dc:source>29000-h.htm</dc:source>

    <meta name='cover' content='item0' />
  </metadata>
    """

        # OPF 2.0 v1.0 specifies to use the
        # Dublin Core Metadata Element Set, Version 1.1
        # http://dublincore.org/documents/2004/12/20/dces/

        dcterms = ElementMaker(nsmap=self.nsmap, namespace=str(NS.dc))

        if dc.publisher:
            self.metadata.append(dcterms.publisher(dc.publisher))
        if dc.rights:
            self.metadata.append(dcterms.rights(dc.rights))

        self.metadata.append(dcterms.identifier(dc.opf_identifier, {'id': 'id'}))

        for count, author in enumerate(dc.authors):
            pretty_name = dc.make_pretty_name(author.name)
            if author.marcrel in {'aut', 'cre'}:
                self.metadata.append(dcterms.creator(pretty_name, {'id': f'author_{count}'}))
            else:
                self.metadata.append(dcterms.contributor(pretty_name, {'id': f'author_{count}'}))
            self.metadata.append(self.opf.meta(author.name,
                                               {'property':'file-as',
                                                'refines': f'#author_{count}'}))
            self.metadata.append(self.opf.meta(author.marcrel,
                                               {'property':'role',
                                                'refines': f'#author_{count}',
                                                'scheme': 'marc:relators'}))


        # replace newlines with /
        title = re.sub(r'\s*[\r\n]+\s*', ' / ', dc.title)
        self.metadata.append(dcterms.title(title))

        for language in dc.languages:
            self.metadata.append(dcterms.language(language.id))

        for subject in dc.subjects:
            self.metadata.append(dcterms.subject(subject.subject))

        if dc.release_date != datetime.date.min:
            self.metadata.append(dcterms.date(
                dc.release_date.isoformat()))

        self.metadata.append(self.opf.meta(
            datetime.datetime.now(gg.UTC()).isoformat(timespec='seconds').replace('+00:00', 'Z'),
            {'property': 'dcterms:modified'}))

        source = dc.source
        if hasattr(options.config, 'FILESDIR'):
            if source.startswith(options.config.FILESDIR):
                source = source[len(options.config.FILESDIR):]
                source = urllib.parse.urljoin(options.config.PGURL, source)

        self.metadata.append(dcterms.source(source))


    def add_coverpage(self, url, id_):
        """ Add a coverpage for ADE and Kindle.        """

        debug("Adding coverpage id: %s url: %s" % (id_, url))

        # register mobipocket style
        self.meta_item('cover', id_)



class Writer(EpubWriter.Writer):
    """ Class that writes epub files. """

    VALIDATOR = 'EPUB_VALIDATOR'

    def remove_coverpage(self, xhtml, url):
        """ Remove coverpage from flow.

        EPUB readers will display the coverpage from the manifest and
        if we don't remove it from flow it will be displayed twice.

        """
        for img in xpath(
                xhtml,
                "//xhtml:img[@src = $url and not(contains(@class, 'x-ebookmaker-important'))]",
                url=url):
            debug("remove_coverpage: dropping <img> %s from flow" % url)
            img.drop_tree()
            return # only the first one though

    @staticmethod
    def html_for_epub3(xhtml):
        """ Convert data-epub attribute to ebub attributes
        """
        for e in xpath(xhtml, "//@*[starts-with(name(), 'data-epub')]/.."):
            for key in e.attrib.keys():
                if key.startswith('data-epub-'):
                    val = e.attrib[key]
                    del e.attrib[key]
                    new_key = getattr(NS.epub, key[10:])
                    e.attrib[new_key] = val

    @staticmethod
    def fix_incompatible_css(sheet):
        """ Strip CSS properties and values that are not EPUB3 compatible.
            Remove "media handheld" rules
        """
        cssclass = re.compile(r'\.(-?[_a-zA-Z]+[_a-zA-Z0-9-]*)')
        for rule in sheet:
            if rule.type == rule.MEDIA_RULE:
                for medium in rule.media:
                    info(f'{medium}')
                    if medium == 'handheld':
                        rule.media.deleteMedium(medium)
                        rule.media.appendMedium(HANDHELD_QUERY)

            if rule.type == rule.STYLE_RULE:
                ruleclasses = list(cssclass.findall(rule.selectorList.selectorText))
                for p in list(rule.style):
                    # Apple books only allows position property in fixed-layout books
                    if p.name == 'position':
                        debug("Dropping property %s" % p.name)
                        rule.style.removeProperty('position')
                        rule.style.removeProperty('left')
                        rule.style.removeProperty('right')
                        rule.style.removeProperty('top')
                        rule.style.removeProperty('bottom')

    def shipout(self, job, parserlist, ncx):
        """ Build the zip file. """

        try:
            ocf = OEBPSContainer(
                os.path.join(os.path.abspath(job.outputdir), job.outputfile),
                ('%d/' % options.ebook if options.ebook else None))

            opf = ContentOPF()

            opf.metadata_item(job.dc)

            # write out parserlist

            for p in parserlist:
                try:
                    ocf.add_bytes(self.url2filename(p.attribs.url), p.serialize(),
                                  p.mediatype())
                    if p.mediatype() == mt.xhtml:
                        opf.spine_item_from_parser(p)
                    else:
                        opf.manifest_item_from_parser(p)
                except Exception as what:
                    error("Could not process file %s: %s" % (p.attribs.url, what))

            # toc


            opf.toc_item('toc.xhtml')
            ocf.add_unicode('toc.xhtml', str(ncx))

            for p in parserlist:
                if 'icon' in p.attribs.rel:
                    cover_url = p.attribs.url
                    break
            else:
                # no  cover items. should not happen
                critical('no cover image available. turn on --generate_cover option')
                cover_url = ''

            #register an ADE cover
            href = ocf.add_image_wrapper(Writer.url2filename(cover_url), 'Cover')
            opf.spine_item(href, mt.xhtml, id_='coverpage-wrapper', first=True)

            opf.rewrite_links(self.url2filename)
            ocf.add_unicode('content.opf', str(opf))

            ocf.commit()

        except Exception as what:
            exception("Error building Epub: %s" % what)
            ocf.rollback()
            raise


    def build(self, job):
        """ Build epub """

        ncx = Toc(job.dc)
        parserlist = []
        css_count = 0
        boilerplate_done = False

        # add CSS parsers
        self.add_external_css(job.spider, None, PRIVATE_CSS, 'pgepub.css')

        try:
            chunker = HTMLChunker.HTMLChunker(version='epub3')
            coverpage_url = None

            # do images early as we need the new dimensions later
            for p in job.spider.parsers:
                if hasattr(p, 'resize_image'):
                    if 'icon' in p.attribs.rel:
                        np = p.resize_image(MAX_IMAGE_SIZE, MAX_COVER_DIMEN)
                        np.id = p.attribs.get('id', 'coverpage')
                        coverpage_url = p.attribs.url
                    elif 'linked_image' in p.attribs.rel:
                        np = p.resize_image(LINKED_IMAGE_SIZE, LINKED_IMAGE_DIMEN)
                        np.id = p.attribs.get('id')
                    else:
                        np = p.resize_image(MAX_IMAGE_SIZE, MAX_IMAGE_DIMEN)
                        np.id = p.attribs.get('id')
                    parserlist.append(np)

            for p in job.spider.parsers:
                if p.mediatype() in OPS_CONTENT_DOCUMENTS:
                    debug("URL: %s" % p.attribs.url)

                    if hasattr(p, 'rst2epub2'):
                        xhtml = p.rst2epub2(job)

                        if options.verbose >= 2:
                            # write html to disk for debugging
                            debugfilename = os.path.join(os.path.abspath(job.outputdir),
                                                         job.outputfile)
                            debugfilename = os.path.splitext(debugfilename)[0] + '.' + \
                                job.maintype + '.debug.html'
                            with open(debugfilename, 'wb') as fp:
                                fp.write(etree.tostring(xhtml, encoding='utf-8'))

                    else:
                        # make a copy so we can mess around
                        p.parse()
                        xhtml = copy.deepcopy(p.xhtml) if hasattr(p, 'xhtml') else None

                    if xhtml is not None:
                    
                        # can't have absolute positions in reflowable EPUB3
                        strip_classes = self.get_classes_with_prop(xhtml, props=('position'))                        
                        strip_classes = strip_classes.intersection(STRIP_CLASSES)
                        if strip_classes:
                            self.strip_pagenumbers(xhtml, strip_classes)

                        if not boilerplate_done:
                            HTMLWriter.Writer.replace_boilerplate(job, xhtml)
                            boilerplate_done = True
                        HTMLWriter.Writer.xhtml_to_html(xhtml)

                        self.html_for_epub3(xhtml)
                        xhtml.make_links_absolute(base_url=p.attribs.url)

                        # build up TOC
                        # has side effects on xhtml
                        ncx.toc += p.make_toc(xhtml)

                        # allows authors to customize css for epub
                        self.add_body_class(xhtml, 'x-ebookmaker')
                        self.add_body_class(xhtml, 'x-ebookmaker-3')

                        self.insert_root_div(xhtml)

                        # strip all links to items not in manifest
                        p.strip_links(xhtml, job.spider.dict_urls_mediatypes())
                        self.strip_links(xhtml, job.spider.dict_urls_mediatypes())

                        self.strip_noepub(xhtml)

                        self.fix_html_image_dimensions(xhtml)
                        if coverpage_url and not hasattr(p.attribs, 'nonlinear'):
                            self.remove_coverpage(xhtml, coverpage_url)

                        # externalize and fix CSS
                        for style in xpath(xhtml, '//xhtml:style'):
                            self.add_external_css(
                                job.spider, xhtml, style.text, "%d.css" % css_count)
                            css_count += 1
                            style.drop_tree()

                        self.add_external_css(job.spider, xhtml, None, 'pgepub.css')

                        self.add_meta_generator(xhtml)

                        debug("Splitting %s ..." % p.attribs.url)
                        chunker.next_id = 0
                        chunker.split(xhtml, p.attribs)
                    else:
                        # parsing xml worked, but it isn't xhtml. so we need to reset mediatype
                        # to something that isn't recognized as content
                        p.attribs.mediatype = parsers.ParserAttributes.HeaderElement('text/xml')
            for p in job.spider.parsers:
                if str(p.attribs.mediatype) == 'text/css':
                    p.parse()
                if hasattr(p, 'sheet') and p.sheet:
                    self.fix_incompatible_css(p.sheet)
                    if job.subtype == '.noimages':
                        p.strip_images()
                    p.rewrite_links(self.url2filename)
                    parserlist.append(p)
                if str(p.attribs.mediatype) in OPS_FONT_TYPES:
                    warning('font file embedded: %s ;  check its license!', p.attribs.url)
                    parserlist.append(p)

            # after splitting html into chunks we have to rewrite all
            # internal links in HTML
            chunker.rewrite_internal_links()
            # also in the TOC
            if not ncx.toc:
                ncx.toc.append([job.spider.parsers[0].attribs.url, 'Start', 1])
            chunker.rewrite_internal_links_toc(ncx.toc)

            # make absolute links zip-filename-compatible
            chunker.rewrite_links(self.url2filename)
            ncx.rewrite_links(self.url2filename)

            # Do away with the chunker and copy all chunks into new parsers.
            # These are fake parsers that never actually parsed anything,
            # we just use them to just hold our data.
            for chunk, attribs in chunker.chunks:
                p = ParserFactory.ParserFactory.get(attribs)
                p.xhtml = chunk
                parserlist.append(p)

            self.shipout(job, parserlist, ncx)

        except Exception as what:
            exception("Error building Epub: %s" % what)
            raise



# -*- coding: utf-8 -*-
# Copyright: This module is put into the public domain.
# Author: Marcello Perathoner <webmaster@gutenberg.org>

"""
Xetex writer for reStructuredText.

This module is more suitable for writing novel-type books than
documentation.

"""

from __future__ import unicode_literals

__docformat__ = 'reStructuredText'

import operator
import os
import re

import six

from docutils import nodes, frontend, transforms
from docutils.writers.html4css1 import SimpleListChecker

from libgutenberg.Logger import error, warning, info, debug
from libgutenberg import DublinCore
from ebookmaker.Version import VERSION

from ebookmaker.mydocutils import writers
from ebookmaker.mydocutils.transforms import parts

XETEX_PREAMBLE = r"""% -*- mode: tex -*- coding: utf-8 -*-
% Converted from RST master
%
\documentclass[a5paper]{book}

\usepackage{polyglossia}
\setotherlanguages{<otherlanguages>}

\usepackage{xltxtra}

\defaultfontfeatures{Scale=MatchLowercase}
\setmainfont{Libertinus Serif}
\setsansfont{Libertinus Sans}
\setmonofont[HyphenChar=None]{DejaVu Sans Mono}

\usepackage{calc}
\usepackage{graphicx}
\usepackage{alltt}
\usepackage{float}

\usepackage{array} % longtable uses array if loaded
\usepackage{longtable}

\usepackage{booktabs} % typographical tables
\newcommand{\otoprule}{\midrule[\heavyrulewidth]}

\usepackage{lettrine} % dropcaps
\usepackage[implicit=false,colorlinks=true,linkcolor=blue]{hyperref}
<hypersetup>
\usepackage[open,openlevel=1]{bookmark}

\tolerance 10000  % dont make overfull boxes
\hbadness 1000    % warn if badness exceeds 1000

\catcode`@=11     % make 'private' LaTeX variables public
\catcode`\^^J=10  % don't let empty lines end paragraphs
\catcode`\^^M=10
\catcode`\"=12    % no electric quotes

\setlength{\textwidth} {\paperwidth  * 7 / 9}
\setlength{\textheight}{\paperheight * 7 / 9}

% round textheight to baselineskip
\small\normalsize
\advance\textheight by .5\baselineskip
\divide\textheight by \baselineskip
\multiply\textheight by \baselineskip
\advance\textheight by \topskip

\setlength{\topmargin}     {\paperheight / 9 - \topskip - \headsep - \headheight - 1in}

\setlength{\evensidemargin}{\paperwidth / 9 - 1in}
\setlength{\oddsidemargin} {\paperwidth / 9 - 1in}

% \setlength{\fboxsep}{0pt} % fbox used for debugging

\let\par=\endgraf % elude macro parameter protection

\begin{document}

\newcommand{\parindentamount}{24pt}
\setlength{\parindent}{\parindentamount}
\setlength{\parskip}{0pt}
\setlength{\parsep}{0pt}
% \setlength{\topsep}{0pt plus6pt}
\setlength{\footnotesep}{0pt}

% relax float restrictions
\renewcommand{\topfraction}{.85}
\renewcommand{\bottomfraction}{.7}
\renewcommand{\textfraction}{.15}
\renewcommand{\floatpagefraction}{.66}
\renewcommand{\dbltopfraction}{.66}
\renewcommand{\dblfloatpagefraction}{.66}
\setcounter{topnumber}{9}
\setcounter{bottomnumber}{9}
\setcounter{totalnumber}{20}
\setcounter{dbltopnumber}{9}

\setcounter{LTchunksize}{10000} % process tables in one chunk

% latex always wants a numeric parameter to \footnotemark and \footnotetext
% hack to get latex to accept any string as footnote label
\def\@xfootnotemark[#1]{%
   \begingroup
      \unrestored@protected@xdef\@thefnmark{#1}%
   \endgroup
   \@footnotemark}

\def\@xfootnotenext[#1]{%
  \begingroup
     \unrestored@protected@xdef\@thefnmark{#1}%
  \endgroup
  \@footnotetext}


% headers

% HACK! to avoid a page break between labels and section title
% standard secpenalty is -300
\@secpenalty = 0

\setcounter{secnumdepth}{-1} % no automatic section numbering
% \setcounter{tocdepth}{1} we don't use auto toc at present

% \def\pgpageno#1{\marginpar[\hfill\fbox{#1}]{\fbox{#1}}}

% \def\pglineno#1{\@mparswitchfalse\marginparsep-24pt\marginpar{#1}}

\def\sd{\dp\strutbox}

\def\pgpageno#1{\strut\vadjust{\kern-\sd\vtop to \sd{\baselineskip\sd\vss\vbox{\hbox to 0pt{\hss\footnotesize(#1)\kern 18pt}\null}}}}

\def\pglineno#1{\strut\vadjust{\kern-\sd\vtop to \sd{\baselineskip\sd\vss\vbox{\hbox to \hsize{\hfill #1\kern 24pt}\null}}}}

\floatstyle{boxed}
\newfloat{topic}{tbp}{loc}
\newfloat{sidebar}{tbp}{los}

\long\def\centerleft#1{%
  {\sbox\@tempboxa{#1}%
  \ifdim \wd\@tempboxa >\hsize
    #1\par
  \else
    {\centering #1\par}
  \fi}}

\setlength{\belowcaptionskip}{\smallskipamount}

\def\raggedright{%
  \let\\\@centercr
  \advance\rightskip\@flushglue
  \parindent\z@}

\def\centering{%
  \let\\\@centercr
  \advance\rightskip\@flushglue\advance\leftskip\@flushglue
  \parindent\z@\parfillskip\z@skip}

\def\raggedleft{%
  \let\\\@centercr
  \advance\leftskip\@flushglue
  \parindent\z@\parfillskip\z@skip}

\renewenvironment{verse}{\parindent\z@}{}

% insert vertical spacing
\def\vs#1{\vspace{#1\baselineskip}}

% one line of verse
\def\vl{\advance\leftskip48pt\hskip-48pt}

% define environments for most of the
% standard building blocks of a book

\def\startenv{%
  \thispagestyle{empty}%
}
\def\endenv{%
}

\newenvironment{container}{}{\par}

<environments>

\newdimen{\tablewidth} % helper

% \tracingpages=1

\frontmatter

\thispagestyle{empty}

"""

XETEX_POSTAMBLE = r"""
\end{document}

% Local Variables:
% mode: tex
% encoding: utf-8
% End:
"""

class InheritTransform (parts.InheritTransform):
    pass


class FootnotesTransform (transforms.Transform):
    """
    Moves footnote bodies near to footnote references.

    Inserts the footnote body immediately after the footnote reference
    in Xetex format.

    """

    default_priority = 717

    def apply (self, **kwargs):
        # reversed () makes footnotes inside of footnotes come out in
        # the right order
        footnote_refs = list (self.document.traverse (nodes.footnote_reference))
        for footnote_reference in reversed (footnote_refs):
            if 'refid' not in footnote_reference:
                # further references to footnote
                continue
            # pull in footnote body
            footnote = self.document.ids [footnote_reference['refid']]
            footnote_section = footnote.parent
            footnote_section.remove (footnote)

            parent = footnote_reference.parent
            index = parent.index (footnote_reference)
            parent.insert (index + 1, footnote)

            # if the footnote is inside a table environment, move the
            # footnote behind the table
            while parent:
                table = parent
                parent = parent.parent
                if isinstance (table, (nodes.table, nodes.title, nodes.footnote)):
                    footnote.parent.remove (footnote)
                    index = parent.index (table)
                    parent.insert (index + 1, footnote)


class Writer (writers.Writer):
    """ A xetex/pdf writer. """

    supported = ('xetex',)
    """Formats this writer supports."""

    output = None
    """Final translated form of `document`."""

    settings_spec = (
        'Xetex-Specific Options',
        None,
        (('Should lists be compacted when possible?',
          ['--compact-lists'],
          {'default': 1,
           'action': 'store_true',
           'validator': frontend.validate_boolean}),
         ('Image Size Function',
          ['--get-image-size'],
          {'default': None,
           'dest': 'get_image_size',
           'metavar': '<python function>'}),
         ))

    config_section = 'XeTeX writer'


    def get_transforms (self):
        tfs = writers.Writer.get_transforms (self)
        return tfs + [parts.TextNodeWrapper, InheritTransform, FootnotesTransform]


    def __init__ (self):
        writers.Writer.__init__ (self)
        self.translator_class = Translator


    def translate (self):
        visitor = self.translator_class (self.document)
        self.document.walkabout (visitor)
        self.output = visitor.astext ()


class MakeTableRules (nodes.SparseNodeVisitor):

    """
    Makes a pass over a table row to output rules.
    """

    def __init__ (self, document, translator):
        nodes.SparseNodeVisitor.__init__ (self, document)
        self.translator = translator

    def unknown_visit (self, node):
        pass

    def visit_entry (self, node):
        """ Called on each table cell. """

        if 'vspan' in node:
            raise nodes.SkipNode

        firstcol = node['column']
        lastcol = firstcol + len (node.colspecs) - 1
        self.translator.cmd (r'\cmidrule{%d-%d}' % (firstcol + 1, lastcol + 1)) # 1-based


class ContentsFilter (nodes.TreeCopyVisitor):
    """ Clean a title for inclusion in a TOC. """

    def visit_citation_reference (self, node):
        raise nodes.SkipNode

    def visit_footnote_reference (self, node):
        raise nodes.SkipNode

    def visit_bullet_list (self, node):
        # skip titles of lesser sections (TOCs are nested lists)
        raise nodes.SkipNode

    def visit_inline (self, node):
        if 'toc-pageref' in node['classes']:
            raise nodes.SkipNode
        self.default_visit (node)

    def visit_newline (self, node):
        self.parent.append (nodes.Text (' '))
        self.default_visit (node)

    def depart_newline (self, node):
        self.default_departure (node)

    def visit_image (self, node):
        if 'alt' in node:
            self.parent.append (nodes.Text (node['alt']))
        raise nodes.SkipNode

    def ignore_node_but_process_children (self, node):
        raise nodes.SkipDeparture

    visit_interpreted = ignore_node_but_process_children
    visit_problematic = ignore_node_but_process_children
    visit_reference = ignore_node_but_process_children
    visit_target = ignore_node_but_process_children


class Translator (writers.Translator):
    """ XeTeX translator """

    # beforeskip, afterskip and font
    # values from latex book.cls
    section_params = (
        (r'\vspace*{50pt}\thispagestyle{empty}',
                            r'\vskip 40pt',  r'\normalfont\Huge\bfseries'),
        (r'\vskip 3.5ex',   r'\vskip 2.3ex', r'\normalfont\Large\bfseries'),
        (r'\vskip 3.25ex',  r'\vskip 1.5ex', r'\normalfont\large\bfseries'),
        (r'\vskip 3.25ex',  r'\vskip 1.5ex', r'\normalfont\normalsize\bfseries'),
        (r'\vskip 3.25ex',  r'\vskip 1.5ex', r'\normalfont\normalsize\bfseries'),
        (r'\vskip 3.25ex',  r'\vskip 1.5ex', r'\normalfont\normalsize\bfseries'),
        )

    # environments that need special treatment, defined in XETEX_PREAMBLE
    page_environments = """
    coverpage frontispiece titlepage verso dedication plainpage
    """.split ()

    # more environments that don't get special treatment
    # but might be useful in post-processing the TeX file
    special_environments = """
    contents foreword preface introduction prologue epilogue glossary
    bibliography index colophon pgfooter pgheader appendix
    """.split ()


    def __init__ (self, document):
        writers.Translator.__init__ (self, document)

        self.forbidden = '' # temporary hold for statements that are forbidden
                            # in environments. eg. footnotes inside tables.

        self.base_dir = os.path.dirname (document.settings.base_url)
        self.last_output_char = '\n'
        self.indent_p = False
        self.used_languages = set (('english', ))
        self.in_float = 0 # are we in a float? (latex doesn't allow nested floats)


    def register_classes (self):
        """ Register classes. """

        # register classes in the order you want them applied!

        # outside block classes

        # mayor structural building blocks
        for i in self.page_environments + self.special_environments:
            self.register_class  ('compound simple', i, '\\begin{%s_env}\n' % i, '\\end{%s_env}\n' % i)

        # simple classes
        # are blocks that contain only inlines

        self.register_class ('simple', 'noindent',     '\\noindent\n',     '')
        self.register_class ('simple', 'left',         '{\\raggedright\n', '\\endgraf}\n')
        self.register_class ('simple', 'center',       '{\\centering\n',   '\\endgraf}\n')
        self.register_class ('simple', 'right',        '{\\raggedleft\n',  '\\endgraf}\n')
        self.register_class ('simple', 'centerleft',   '{\\centerleft{',   '}}\n')

        # inline classes
        # be careful not to insert spurious spaces here

        self.register_class ('inline', 'superscript', r'\textsuperscript{', r'}')
        self.register_class ('inline', 'subscript',   r'\textsubscript{',   r'}')
        self.register_class ('inline', 'italics',     r'{\itshape{',        r'}}')
        self.register_class ('inline', 'bold',        r'{\bfseries{',       r'}}')
        self.register_class ('inline', 'monospaced',  r'{\ttfamily{',       r'}}')
        self.register_class ('inline', 'small-caps',  r'{\scshape{',        r'}}')
        self.register_class ('inline', 'normal',      r'{\upshape{',        r'}}')
        self.register_class ('inline', 'antiqua',     r'{\upshape{',        r'}}')
        self.register_class ('inline', 'gesperrt',    r'{\addfontfeature{LetterSpace=20.0}{',  r'}}')
        self.register_class ('inline', 'larger',      r'{\addfontfeature{Scale=1.2}{',         r'}}')
        self.register_class ('inline', 'smaller',     r'{\addfontfeature{Scale=0.8}{',         r'}}')

        self.register_class ('inline', 'dropspan',    r'{', r'}')

        # The height of a TeX box is just the height of the character,
        # while the height of an HTML line box is the height of the
        # character + internal leading.  Adding \strut is a sorry fix
        # for the problem of paragraphs with different font sizes.  It
        # will work only for spans of text not longer than 2 lines.  A
        # real fix must insert \strut before and/or after every word
        # in the bigger text.  A fix involving \baselineskip or
        # \lineskip and \lineskiplimit will not work because those
        # values will be reset before the paragraph builder gets
        # exercised.

        self.register_class ('inline', 'xx-large',    r'{\Huge\strut{',         r'}}')
        self.register_class ('inline', 'x-large',     r'{\LARGE\strut{',        r'}}')
        self.register_class ('inline', 'large',       r'{\Large\strut{',        r'}}')
        self.register_class ('inline', 'medium',      r'{\normalsize{ ',   r'}}')
        self.register_class ('inline', 'small',       r'{\footnotesize{', r'}}')
        self.register_class ('inline', 'x-small',     r'{\scriptsize{',   r'}}')
        self.register_class ('inline', 'xx-small',    r'{\tiny{',         r'}}')

        self.register_class ('inline', 'red',         r'{\addfontfeature{Color=FF0000}{',      r'}}')
        self.register_class ('inline', 'green',       r'{\addfontfeature{Color=00FF00}{',      r'}}')
        self.register_class ('inline', 'blue',        r'{\addfontfeature{Color=0000FF}{',      r'}}')
        self.register_class ('inline', 'yellow',      r'{\addfontfeature{Color=FFFF00}{',      r'}}')
        self.register_class ('inline', 'white',       r'{\addfontfeature{Color=FFFFFF}{',      r'}}')
        self.register_class ('inline', 'gray',        r'{\addfontfeature{Color=808080}{',      r'}}')
        self.register_class ('inline', 'black',       r'{\addfontfeature{Color=000000}{',      r'}}')


    def cmd (self, cmds):
        """ Output tex commands. """

        if isinstance (cmds, six.string_types):
            cmds = [cmds]

        for c in cmds:
            if c:
                self.context.append (c)
                self.last_output_char = c[-1]

    def text (self, text):
        """ Output text. """

        if text:
            # self.output_noindent ()
            self.context.append (text)
            self.last_output_char = text[-1]


    def comment (self, text):
        """ Output tex comment. """
        self.context.append ('%% %s\n' % text)
        self.last_output_char = '\n'
        self.src_sp ()

    def nl (self):
        if self.last_output_char != '\n':
            self.context.append ('\n')
            self.last_output_char = '\n'

    def noindent (self, b = True):
        """ Don't indent following text. """
        self.indent_p = not b

    def output_sp (self):
        """ Output spacing. """
        if self.vspace == 1999: # magic number to eat all space
            self.vspace = 0
        if self.vspace:
            self.nl ()
            self.cmd ('\\vs{%d}\n' % self.vspace)
            self.vspace = 0
            self.noindent ()

    def output_src_sp (self):
        """ Output source spacing. """
        if self.src_vspace == 1999: # magic number to eat all space
            self.src_vspace = 0
        if self.src_vspace:
            self.nl ()
            self.context.append ('\n' * self.src_vspace)
            self.src_vspace = 0

    def output_noindent (self):
        if not self.indent_p:
            self.context.append ('{\\noindent}\n')
            self.indent_p = True

    def ta (self, indent, text):
        """ Tabulate text to indent position. """
        self.cmd (r'\hspace{%dem}\=\kill\+' % indent)
        self.text (text)

    def push (self):
        """ Push environment. """
        self.context.append ('{')
        self.last_output_char = '{'

    def pop (self):
        """ Pop environment. """
        self.context.append ('}')
        self.last_output_char = '}'

    def begin (self, name, param = None):
        """ Push environment. """
        self.output_src_sp ()
        self.environments.append (name)
        self.cmd ('\\begin{%s}%s\n' % (name, param or ''))

    def end (self):
        """ Pop environment. """
        name = self.environments.pop ()
        self.nl ()
        self.cmd ('\\end{%s}\n' % name)

    def filter_title (self, node):
        """ Return a copy of a title, with references, images, etc. removed."""
        visitor = ContentsFilter (self.document)
        node.walkabout (visitor)
        return visitor.get_tree_copy ()

    def make_relative (self, uri):
        # Use relative paths in the TeX file because we copied the
        # image files to a local disk. (TeX cannot pull files from the
        # network.)

        # print ("make_relative: %s from %s" % (uri, self.base_dir))
        uri = os.path.relpath (uri, self.base_dir)
        # print ("made_relative: %s" % uri)

        return uri

    def get_align (self, node):
        return node.colspecs[0].get ('align', 'justify')[0] # .upper ()

    def indent (self, by = 1):
        """ Indent text. """
        self.cmd (r'\advance\leftskip\parindentamount{}') #  % by)

    def rindent (self, by = 1):
        """ Indent text on the right side. """
        self.cmd (r'\advance\rightskip\parindentamount{}') #  % by)

    def bq (self, by = 1):
        self.indent (by)
        self.rindent (by)
        self.noindent ()

    # pylint: disable=C0111
    # pylint: disable=W0613

    translate_map = {
        ord ('#'):  r'\#',
        ord ('$'):  r'\$',
        ord ('%'):  r'\%',
        ord ('&'):  r'\&',
        ord ('~'):  r'\textasciitilde{}',
        ord ('_'):  r'\_',
        ord ('^'):  r'\textasciicircum{}',
        ord ('\\'): r'\textbackslash{}',
        ord ('{'):  r'\{',
        ord ('}'):  r'\}',
        ord ('['):  r'{[}',
        ord (']'):  r'{]}',
        0x00ad:     r'\-', # soft hyphen
    }

    translate_map_literal = {}
    translate_map_literal.update (translate_map)
    translate_map_literal.update ({
        0x0d:       '\\\\n',
    })

    def preamble (self):
        """ Inserts xetex preamble. """

        hs = ['\\hypersetup{pdfcreator={Ebookmaker %s}}' % VERSION]
        if hasattr (self.document, 'meta_block'):
            for name, values in self.document.meta_block.items ():
                if name.lower () == 'dc.title':
                    content = values[0]
                    hs.append ('\\hypersetup{pdftitle={%s}}' % self.encode (content))
                elif name.lower () == 'dc.creator':
                    content = DublinCore.DublinCore.strunk (values)
                    hs.append ('\\hypersetup{pdfauthor={%s}}' % self.encode (content))
                elif name.lower () == 'dc.subject':
                    content = DublinCore.DublinCore.strunk (values)
                    hs.append ('\\hypersetup{pdfsubject={%s}}' % self.encode (content))

        envs = []
        for i in self.page_environments:
            envs.append (r'\newenvironment{%s_env}{\startenv}{\endenv}' % i)
        for i in self.special_environments:
            envs.append (r'\newenvironment{%s_env}{}{}' % i)

        preamble = XETEX_PREAMBLE
        preamble = preamble.replace ('<environments>', '\n'.join (envs))
        preamble = preamble.replace ('<otherlanguages>', ','.join (self.used_languages))
        return preamble.replace ('<hypersetup>', '\n'.join (hs))

    def postamble (self):
        """ Inserts xetex postamble. """
        return XETEX_POSTAMBLE

    def write_labels (self, node):
        """ Write labels for all ids and the refid of `node` """

        ids = node['ids']
        refid = node.get ('refid')
        if refid is not None:
            ids.append (refid)

        for id_ in ids:
            self.cmd ('\\label{%s}%%\n' % id_)
            self.cmd ('\\hypertarget{%s}{}%%\n' % id_)

    def write_pdf_bookmark (self, title, action_option, level = None):
        if level is None:
            level = self.section_level
        options = [ 'level=%d' % level, action_option ]
        self.cmd ('\\bookmark[%s]{%s}\n' % (','.join (options), self.encode (title)))

    def latex_units (self, length_str, reference_length = ''):
        """ Convert rst units to LaTeX units. """

        if length_str == 'image':
            return reference_length

        match = re.match (r'(\d*\.?\d*)\s*(\S*)', length_str)
        if not match:
            return length_str

        value, unit = match.groups ()
        # no unit or "DTP" points (called 'bp' in TeX):
        if unit in ('', 'pt'):
            length_str = '%sbp' % value

        # percentage: relate to current line width
        elif unit == '%':
            length_str = '%.3f%s' % (float (value) / 100.0, reference_length)

        return length_str

    def latex_floats (self, values):
        if 'none' in values:
            # 'H' must be only option, as per float package doc
            return 'H'
        lf = ''
        for v in values:
            lf += { 'here': 'h!', 'top': 't', 'bottom': 'b', 'page': 'p' }[v]
        if lf == 'h!':
            lf = 'h!t'
        return lf

    def begin_float (self, node, type_):
        """ Potentially open a latex floating environment. """

        ltxfl = self.latex_floats (node['float'])

        # latex doesn't support nested floating environments
        node.do_float = (ltxfl != 'H') and (self.in_float == 0)

        if node.do_float:
            # if we float we must save the accumulated space and remove it
            # or it will show up at the top inside the float
            node.saved_vspace = self.vspace
            self.vspace = 0

            # if we float away, make the next paragraph indent
            # note: we don't indent paras after figures any more
            #for para in node.traverse (nodes.paragraph, include_self = False, siblings = True):
            #    classes = para['classes']
            #    classes.append ('pnext')
            #    if 'pfirst' in classes:
            #        classes.remove ('pfirst')

            self.in_float += 1
            self.begin (type_, '[%s]' % ltxfl)
        else:
            self.begin ('container')

    def end_float (self, node):
        self.end ()
        if node.do_float:
            self.in_float -= 1
            # FIXME: floats insert their own breaks. Do we still needs this?
            self.vspace = node.saved_vspace
            # self.vspace = 0

    def encode (self, text):
        return text.translate (self.translate_map)

    def latex_language (self, node):
        """ language hack """

        for c in node['classes']:
            if c.startswith ('language-'):
                language = c[9:]
                language = DublinCore.DublinCore.language_map.get (
                    language, default='en').lower ()
                self.used_languages.add (language)
                return language
        return None

    # begin visitor functions

    def visit_outer (self, node):
        # self.comment ("visit_outer: %s" % node.__class__.__name__)

        if node.type in ('compound', 'simple'):
            if not isinstance (node, (nodes.footnote, nodes.label)):
                # often footnotes are moved inside of inline text
                # a newline in the source would then insert a spurious space
                self.output_src_sp ()

            language = self.latex_language (node)
            if language:
                self.begin (language)

        # transition and image
        if node.type == 'empty':
            if 'margin' in node['classes']:
                self.sp ()
            self.output_sp ()

    def visit_inner (self, node):
        # self.comment ("visit_inner: %s" % node.__class__.__name__)

        # must be here because it belongs inside a potential float
        if node.type != 'text' and 'margin' in node['classes']:
            self.sp ()

        # There can be only inline nodes inside a simple node, and
        # inline nodes have no margins, so we can safely output
        # vertical space here.
        if node.type == 'simple':
            self.output_sp ()
            self.output_noindent ()

        if node.type == 'inline':
            language = self.latex_language (node)
            if language:
                self.cmd (r'\text%s{' % language)
            self.cmd (self.get_prefix (node.type, node['classes']))

        elif node.type != 'text':
            prefix = self.get_prefix (node.type, node['classes'])
            if prefix:
                self.nl ()
                self.cmd (prefix)

    def depart_inner (self, node):
        if node.type == 'inline':
            self.cmd (self.get_suffix (node.type, node['classes']))
            if self.latex_language (node):
                self.cmd (r'}')
        elif node.type != 'text':
            suffix = self.get_suffix (node.type, node['classes'])
            if suffix:
                self.nl ()
                self.cmd (suffix)
        if node.type != 'text' and 'margin' in node['classes']:
            self.sp ()

        # self.comment ("departed_inner: %s" % node.__class__.__name__)

    def depart_outer (self, node):
        if node.type == 'empty':
            if 'margin' in node['classes']:
                self.sp ()

        if node.type in ('compound', 'simple'):
            if self.latex_language (node):
                self.end ()
            self.src_sp ()

        # self.comment ("departed_outer: %s" % node.__class__.__name__)


    def visit_Text (self, node):
        text = node.astext ()
        if self.in_literal:
            text = text.translate (self.translate_map_literal)
        else:
            text = text.translate (self.translate_map)

        if hasattr (node, 'attributes') and 'white-space-pre-line' in node.attributes['classes']:
            for line in text.splitlines (True):
                self.text (line)
                if '\n' in line:
                    self.cmd (' \\\\\n')
        else:
            self.text (text)

    def depart_Text (self, node):
        pass

    def visit_inline (self, node):
        if 'toc-pageref' in node['classes']:
            self.cmd (r'\leaders\hbox to 1em{\hss.\hss}\hfill{}')
        if 'dropcap' in node['classes']:
            self.lettrine (node)

    def depart_inline (self, node):
        if 'dropcap' in node['classes']:
            self.cmd (r'}')

    # start docinfo elements (parse only for now)

    def visit_docinfo (self, node):
        pass

    def depart_docinfo (self, node):
        pass

    def visit_authors (self, node):
        pass

    def depart_authors (self, node):
        pass

    def visit_field (self, node):
        pass

    def depart_field (self, node):
        pass

    def visit_field_name (self, node):
        self.field_name = node.astext ().lower ().replace (' ', '_')
        raise nodes.SkipNode

    def depart_field_name (self, node):
        pass

    def visit_field_body (self, node, name = None):
        # name either from element or stored by <field_name>
        self.context = self.docinfo[name or self.field_name]

    def depart_field_body (self, node):
        self.context = self.body

    def visit_field_list (self, node):
        self.push ()
        self.indent (2)

    def depart_field_list (self, node):
        self.pop ()

    # start admonitions

    def visit_admonition (self, node, name = None):
        self.sp ()
        self.push ()
        self.bq ()
        if name:
            self.output_sp ()
            self.cmd ('\\textbf{')
            self.text (name)
            self.cmd ('}\\par\\nopagebreak\n\n')

    def depart_admonition (self, node):
        self.pop ()
        self.sp ()

    def visit_admonition_title (self, node):
        pass

    def depart_admonition_title (self, node):
        self.cmd (r'\par')
        self.src_sp ()
        self.sp ()

    # start definition lists

    def visit_definition_list (self, node):
        self.begin ('description')

    def depart_definition_list (self, node):
        self.end ()

    def visit_definition_list_item (self, node):
        pass

    def depart_definition_list_item (self, node):
        pass

    def visit_term (self, node):
        self.cmd ('\\item[')

    def depart_term (self, node):
        self.cmd ('] ')

    def visit_classifier (self, node):
        pass

    def depart_classifier (self, node):
        pass

    def visit_definition (self, node):
        pass

    def depart_definition (self, node):
        pass

    # start option lists

    def visit_option_list (self, node):
        pass

    def depart_option_list (self, node):
        pass

    def visit_option_list_item (self, node):
        pass

    def depart_option_list_item (self, node):
        pass

    def visit_option_group (self, node):
        pass

    def depart_option_group (self, node):
        pass

    def visit_option (self, node):
        pass

    def depart_option (self, node):
        if 'last' not in node['classes']:
            self.text (', ')

    def visit_option_string (self, node):
        pass

    def depart_option_string (self, node):
        pass

    def visit_option_argument (self, node):
        self.text (node.get ('delimiter', ' '))

    def depart_option_argument (self, node):
        pass

    def visit_description (self, node):
        pass

    def depart_description (self, node):
        pass

    # lists

    def check_simple_list (self, node):
        """Check for a simple list that can be rendered compactly."""
        try:
            node.walk (SimpleListChecker (self.document))
            return True
        except nodes.NodeFound:
            return False

    def is_compactable (self, node):
        return ('compact' in node['classes']
                or (self.settings.compact_lists
                    and 'open' not in node['classes']
                    and (# self.compact_simple or
                         # self.topic_classes == ['contents'] or
                         self.check_simple_list (node))))

    def list_start (self, node):
        self.begin ('itemize')
        self.list_enumerator_stack.append (writers.ListEnumerator (node, 'utf-8'))

    def list_end (self, node):
        self.list_enumerator_stack.pop ()
        self.end ()

    def visit_list_item (self, node):
        if 'toc-entry' in node['classes']:
            self.cmd (r'\item[] ')
        else:
            self.cmd (r'\item[%s] ' % self.list_enumerator_stack[-1].get_next ())

    def depart_list_item (self, node):
        pass

    def visit_bullet_list (self, node):
        self.list_start (node)

    def depart_bullet_list (self, node):
        self.list_end (node)

    def visit_enumerated_list (self, node):
        self.list_start (node)

    def depart_enumerated_list (self, node):
        self.list_end (node)

    # end lists

    def visit_block_quote (self, node):
        self.set_first_last (node)
        self.sp ()
        self.push ()
        self.bq ()

    def depart_block_quote (self, node):
        self.pop ()
        self.sp ()

    def visit_comment (self, node):
        for line in node.astext ().splitlines ():
            self.comment (line)
        raise nodes.SkipNode

    def visit_container (self, node):
        self.begin ('container')

    def depart_container (self, node):
        self.end ()
        self.noindent ()

    def lettrine (self, node):
        # node is inline or image

        options = []
        if 'lines' in node:
            options.append ('lines=%d' % node['lines'])
        if 'indents' in node:
            indents = node['indents'].split ()
            if len (indents) < 2:
                indents.append ('0.5em')
            self.cmd (r'\dimen0=%s\dimen1=%s\advance\dimen1-\dimen0'
                      % (indents[0], indents[1]))
            options.append (r'findent=\dimen0')
            options.append (r'nindent=\dimen1')
        if 'raise' in node:
            options.append ('lraise=%f' % node['raise'])

        uri = None
        if isinstance (node, nodes.image):
            uri = node.get ('image', '')
        if uri:
            options.append ('image')
            self.cmd (r'\clubpenalty\@M\lettrine[%s]{%s}' %
                      (','.join (options), self.make_relative (uri)))
        else:
            self.cmd (r'\clubpenalty\@M\lettrine[%s]{' % ','.join (options))

    def visit_compound (self, node):
        pass

    def depart_compound (self, node):
        pass

    def visit_decoration (self, node):
        pass

    def depart_decoration (self, node):
        pass

    def visit_doctest_block (self, node):
        self.visit_literal_block (node)

    def depart_doctest_block (self, node):
        self.depart_literal_block (node)

    def visit_document (self, node):
        pass

    def depart_document (self, node):
        # extract bookmarks from the various 'contents' sections
        for container in node.traverse (nodes.container):
            if set (('contents', 'lof', 'lot')).intersection (container['classes']):
                if isinstance (container.parent[0], nodes.title):
                    title = self.filter_title (container.parent[0]).astext ()
                else:
                    title = 'Section'
                self.cmd ('\\bookmark[level=1,page=1]{%s}\n' % self.encode (title))

                for item in container.traverse (nodes.list_item):
                    text = self.filter_title (item).astext ()
                    self.write_pdf_bookmark (text, 'dest=%s' % item['refid'], item['level'])

        # bookmarks from original page numbers
        page_targets = [target for target in node.traverse (nodes.target)
                        if 'pageno' in target['classes']]
        if page_targets:
            self.cmd ('\\bookmark[level=1,page=1]{Pages In the Original}\n')
            for target in page_targets:
                title = target['html_attributes']['title']
                self.write_pdf_bookmark (title, 'dest=%s' % target['ids'][0], 2)

        # bookmarks need to land on a page
        self.cmd ('\\hbox{}')


    def visit_footer (self, node):
        self.document.reporter.warning (
            'footer not supported', base_node = node)

    def depart_footer (self, node):
        pass

    # footnotes, citations, labels

    def visit_label (self, node):
        # footnote and citation
        self.cmd ('\\footnotetext[')

    def depart_label (self, node):
        self.cmd (']')
        self.push ()
        self.sp (0)
        self.src_sp (0)

    def visit_footnote (self, node):
        pass

    def depart_footnote (self, node):
        self.nl ()
        self.pop ()

    def visit_footnote_reference (self, node):
        # we use \footnotemark in combination with \footnotetext to
        # permit footnotes even in forbidden modes
        self.cmd ('%\n\\footnotemark[')

    def depart_footnote_reference (self, node):
        self.cmd (']')

    def visit_citation (self, node):
        self.visit_footnote (node)

    def depart_citation (self, node):
        self.depart_footnote (node)

    def visit_citation_reference (self, node):
        self.cmd (r'\cite{')

    def depart_citation_reference (self, node):
        self.cmd (r'}')

    # end footnotes

    # references and targets

    def visit_reference (self, node):
        # we don't support extrnal refs
        if 'refuri' in node:
            href = None
        # internal reference
        elif 'refid' in node:
            href = node['refid']
        elif 'refname' in node:
            href = self.document.nameids[node['refname']]
        else:
            raise AssertionError ('Unknown reference.')
        if href:
            self.cmd ('\\hyperlink{%s}' % href)
        self.push ()

    def depart_reference (self, node):
        self.pop ()

    def visit_target (self, node):
        if 'pageno' in node['classes'] or 'lineno' in node['classes']:
            # self.output_noindent ()
            what = 'page' if 'pageno' in node['classes'] else 'line'
            if 'invisible' not in node['classes']:
                pageno = node['html_attributes']['title']
                cmd  = '\\pg%sno{%s}%%\n' % (what, pageno)
                if 'figure' in self.environments:
                    self.forbidden += cmd
                else:
                    self.cmd (cmd)
            # self.output_sp ()
            self.cmd (r'\raisebox{\ht\strutbox}')
            self.push ()

        refid = node.get ('refid')
        if refid is not None:
            # id has been moved to next_node
            # by PropagateTargets Transform in transforms/references.py

            # titles, figures, tables are handled in their visit_* method
            next_node = self.document.ids[refid]
            if not isinstance (next_node, (nodes.section, nodes.figure, nodes.table)):
                self.write_labels (node)
        else:
            self.write_labels (node)

        self.push ()

    def depart_target(self, node):
        if 'pageno' in node['classes'] or 'lineno' in node['classes']:
            self.pop ()
        self.pop ()

    # end references and targets

    def visit_generated (self, node):
        pass

    def depart_generated (self, node):
        pass

    def visit_header (self, node):
        self.document.reporter.warning (
            'header not supported', base_node = node)

    def depart_header (self, node):
        pass

    def visit_figure (self, node):
        self.forbidden = ''
        self.begin_float (node, 'figure')
        self.write_labels (node)

        if node.attributes.get ('align') in ('center', 'right'):
            self.cmd ('\\noindent\\hfill{}')

        if 'width' in node: # figwidth
            self.begin ('minipage', '{%s}' % self.latex_units (node['width'], '\\textwidth'))

    def depart_figure (self, node):
        if 'width' in node: # figwidth
            self.end ()

        if node.attributes.get ('align') == 'center':
            self.cmd ('\\hfill{}')

        self.end_float (node)
        self.cmd (self.forbidden)
        self.forbidden = ''

    image_align_codes = {
        # for inline images: by default latex aligns the bottom.
        'bottom': r'%s',
        'middle': r'\raisebox{-0.5\height}{%s}',
        'top':    r'\raisebox{-\height}{%s}',
        # block images:
        'center': r'{\hbox to \hsize{\hfill%s\hfill}}',
        'left':   r'{\hbox to \hsize{%s\hfill}}',
        'right':  r'{\hbox to \hsize{\hfill%s}}',
        }

    def visit_image (self, node):
        if 'dropcap' in node['classes']:
            self.lettrine (node)
            return

        options = []
        if 'height' in node:
            options.append ('height=%s' % self.latex_units (node['height']))
        if 'scale' in node:
            options.append ('scale=%f'  % (node['scale'] / 100.0))
        if 'width' in node:
            options.append ('width=%s'  % self.latex_units (node['width'], '\\textwidth'))

        options = (options and '[%s,keepaspectratio=true]' % (','.join (options))) or ''

        command = '\\includegraphics%s{%s}' % (options, self.make_relative (node['uri']))

        align = node.attributes.get ('align')
        if align in self.image_align_codes:
            command = self.image_align_codes[align] % command

        self.cmd (command)

    def depart_image (self, node):
        pass

    def visit_caption (self, node):
        # figure caption
        pass

    def depart_caption (self, node):
        pass

    def visit_legend (self, node):
        # figure legend
        pass

    def depart_legend (self, node):
        pass

    def visit_line_block (self, node):
        if not isinstance (node.parent, nodes.line_block):
            # outermost line block
            self.sp ()
            self.output_sp ()
            self.begin ('verse')
        else:
            self.push ()
            self.indent ()

    def depart_line_block (self, node):
        if not isinstance (node.parent, nodes.line_block):
            # outermost line block
            self.end ()
            self.sp ()
        else:
            self.pop ()

    def visit_line (self, node):
        if len (node.astext ()) == 0:
            # empty lines must use \vspace or latex
            # will complain about no line to end
            self.sp ()
            self.output_sp ()
        else:
            self.push ()
            self.cmd (r'\vl ')

    def depart_line (self, node):
        if len (node.astext ()) != 0:
            self.cmd ('\\par')
            self.pop ()
            self.nl ()

        if node is not node.parent[-1]: # not last
            self.src_sp (0)

    def visit_literal_block (self, node):
        self.sp ()
        self.output_sp () # vspace inside of tt block inserts a spurious newline
        self.begin ('alltt')
        self.in_literal += 1

    def depart_literal_block (self, node):
        self.in_literal -= 1
        self.end ()
        self.sp ()

    #
    #
    #

    def visit_paragraph (self, node):
        pass
        # self.output_sp ()
        # self.output_src_sp ()

    def depart_paragraph (self, node):
        # The LaTeX `\footnote` macro appends a strut to the footnote
        # text to ensure correct spacing if the last line of the
        # footnote text has no descenders. If we end the footnote with
        # a \par the strut will be inserted after the \par and will
        # yield an empty line.  The same is true for table cells.
        if isinstance (node.parent, (nodes.footnote, nodes.entry)):
            if not node.next_node (descend = 0, siblings = 1):
                return
        self.cmd (r'\par')
        self.src_sp ()
        if 'white-space-pre-line' in node['classes']:
            self.output_src_sp ()
            self.cmd ('\\vspace{1em}')
            self.src_sp ()

    def visit_raw (self, node):
        if 'tex' in node.get ('format', '').split():
            self.cmd (node.astext ())

        # ignore other raw formats
        raise nodes.SkipNode

    def visit_substitution_definition (self, node):
        """Internal only."""
        raise nodes.SkipNode

    def visit_substitution_reference (self, node):
        self.document.reporter.warning ('"substitution_reference" not supported',
                base_node=node)

    def visit_system_message (self, node):
        self.sp ()
        self.push ()
        self.bq ()
        line = ', line %s' % node['line'] if 'line' in node else ''
        self.text ('"System Message: %s/%s (%s:%s)"'
                  % (node['type'], node['level'], node['source'], line))

    def depart_system_message (self, node):
        self.pop ()
        self.sp ()

    # tables

    def begin_tabular (self, table):
        """ Begin longtable environment, add columns spec. """

        if self.header_sent:
            return

        colspec = ''
        if 'colspec' in table:
            colspec = table['colspec']
        else:
            colspecs = list (table.traverse (nodes.colspec))
            colspec = 'l' * len (colspecs)
            ## for c in colspecs:
            ##     colspec += { 'top'    : 'p' ,
            ##                  'middle' : 'm' ,
            ##                  'bottom' : 'b' }[c.get ('valign', 'middle')]
            ##     colspec += r'{%.03f\tablewidth}' % c['relative_width']


        self.cmd (r'\setlength{\tablewidth}{%s - \tabcolsep * 2 * %d}' % (
            self.latex_units (table.get ('width', '100%'), '\\linewidth'), len (colspecs)))

        self.src_sp ()
        self.begin ('longtable', r'{%s}' % (colspec))

        self.header_sent = True

    def visit_table (self, node):
        pass_1 = writers.TablePass1 (self.document)
        node.walk (pass_1)
        rows = pass_1.rows ()
        cols = pass_1.cols ()

        self.header_sent = False

        self.begin_float (node, 'table')
        self.write_labels (node)
        self.cmd ('\\footnotesize\n')

    def depart_table (self, node):
        self.end_float (node)

    def visit_table_title (self, node):
        self.sp ()

    def depart_table_title (self, node):
        self.cmd ('\\par\n')
        self.cmd ('\\penalty300\n')
        self.src_sp ()
        self.sp ()

    def visit_tgroup (self, node):
        pass

    def depart_tgroup (self, node):
        pass

    def visit_colspec (self, node):
        pass

    def depart_colspec (self, node):
        pass

    def visit_thead (self, node):
        table = node.parent.parent
        self.begin_tabular (table)
        self.set_first_last (node) # mark first row of head
        if 'table' in table['hrules']:
            self.cmd ('\\toprule\n')

    def depart_thead (self, node):
        pass

    def visit_tbody (self, node):
        table = node.parent.parent
        self.begin_tabular (table)
        self.set_first_last (node) # mark first row of body
        if 'table' in table['hrules']:
            self.cmd ('\\otoprule\n')

    def depart_tbody (self, node):
        table = node.parent.parent
        if 'table' in table['hrules']:
            self.cmd ('\\bottomrule\n')
        self.end ()

    def visit_row (self, node):
        table = node.parent.parent.parent
        self.set_first_last (node) # mark first and last cell
        if 'first' not in node['classes']:
            if 'rows' in table['hrules']:
                make_rules = MakeTableRules (self.document, self)
                node.walk (make_rules)
                self.cmd ('\n')

    def depart_row (self, node):
        self.cmd ('\\tabularnewline\n')

    def visit_entry (self, node):

        if 'vspan' in node:
            self.cmd (' & ' )
            raise nodes.SkipNode

        self.sp (0)

        sum_width = sum (map (operator.itemgetter ('relative_width'), node.colspecs))

        cols = node.get ('morecols', 0)
        if cols:
            self.cmd (r'\setlength{\dimen0}{%.03f\tablewidth + \tabcolsep * %d * 2}' % (sum_width, cols))
        else:
            self.cmd (r'\setlength{\dimen0}{%.03f\tablewidth}' % sum_width)

        parboxoptions =  { 'top'    : '[t]',
                           'middle' : '',
                           'bottom' : '[b]',
                           }[node.colspecs[0].get ('valign', 'middle')]

        rows = node.get ('morerows', 0) + 1
        if rows > 1:
            self.cmd (r'\setlength{\dimen1}{\ht\@arstrutbox * %d}' % rows)
            self.cmd (r'\addtolength{\dimen1}{\dp\@arstrutbox * %d}' % (rows - 1))
            if 'rows' in node.parent.parent.parent.parent['hrules']:
                self.cmd (r'\addtolength{\dimen1}{(\aboverulesep + \belowrulesep) * %d}' % (rows - 1))

            parboxoptions = { 'top'    : r'[t][\dimen1][t]',
                              'middle' : r'[c][\dimen1][c]',
                              'bottom' : r'[b][\dimen1][b]',
                              }[node.colspecs[0].get ('valign', 'middle')]

        self.cmd (r'\setbox0\vbox')
        self.push ()

        self.cmd (r'\hsize\dimen0')

        ## self.cmd (r'\fbox')
        ## self.push ()
        self.cmd (r'\parbox%s{\dimen0}' % parboxoptions)
        self.push ()

        self.cmd ({ 'left'   : '\\raggedright',
                    'right'  : '\\raggedleft',
                    'center' : '\\centering',
                    'justify': '' }[node.colspecs[0].get ('align', 'justify')])

        if isinstance (node.parent.parent, nodes.thead):
            self.cmd ('\\bfseries')

        self.cmd (r'\setlength{\parskip}{1em}\noindent')
        ## self.cmd (r'\fontspec[Numbers=Lining]{Linux Libertine O}')
        self.cmd ('\\@arstrut\n')

    def depart_entry (self, node):
        self.cmd (r'\@arstrut')

        self.pop () # parbox
        ## self.pop () # fbox
        self.pop () # setbox0

        cols = node.get ('morecols', 0)
        if cols:
            self.cmd (r'\setlength{\dimen0}{%.03f\tablewidth}' % node.colspecs[0]['relative_width'])
            self.cmd (r'\wd0\dimen0')

        rows = node.get ('morerows', 0) + 1
        if rows > 1:
            self.cmd (r'\ht0\ht\@arstrutbox\dp0 0pt')

        self.cmd (r'\box0')

        self.cmd (' &' * node.get ('morecols', 0))
        if 'last' not in node['classes']:  # not last cell in row
            self.cmd (' & ')


    # end tables

    # Structural Elements and Titles

    def visit_document_title (self, node):
        pass

    def depart_document_title (self, node):
        self.cmd (r'\par')
        self.src_sp ()
        if 'with-subtitle' not in node['classes']:
            self.sp (2)

    def visit_document_subtitle (self, node):
        self.sp (1)

    def depart_document_subtitle (self, node):
        self.cmd (r'\par')
        self.src_sp ()
        self.sp (2)

    def visit_section (self, node):
        self.section_level += 1

    def depart_section (self, node):
        self.section_level -= 1

    def visit_section_title (self, node):
        node.eff_level = min (self.section_level, 6) - 1 # 0-based
        beforeskip, afterskip, font = self.section_params [node.eff_level]
        section = node.parent

        self.output_sp ()
        # prefer a page break *before* hyperlink labels
        # avoids that the link goes to the previous page
        self.cmd ('\\penalty-300%\n')
        self.write_labels (section)

        self.cmd ('%s\n{\\noindent\\interlinepenalty \\@M%s%%\n' % (beforeskip, font))


    def depart_section_title (self, node):
        beforeskip, afterskip, fontsize = self.section_params [node.eff_level]
        self.cmd ('}\n\n\\unskip%s\\nobreak\n' % afterskip)
        self.noindent ()

    def visit_section_subtitle (self, node):
        self.sp (1)

    def depart_section_subtitle (self, node):
        self.cmd (r'\par\nobreak')
        self.src_sp ()
        self.sp (2)
        self.noindent ()

    def visit_rubric (self, node):
        self.sp ()

    def depart_rubric (self, node):
        self.cmd (r'\par')
        self.src_sp ()
        self.sp ()

    def visit_topic (self, node):
        self.begin_float (node, 'topic')

    def depart_topic (self, node):
        self.end_float (node)

    def visit_topic_title (self, node):
        pass

    def depart_topic_title (self, node):
        self.cmd (r'\par')
        self.src_sp ()
        self.sp ()

    def visit_sidebar (self, node):
        self.begin_float (node, 'sidebar')

    def depart_sidebar (self, node):
        self.end_float (node)

    def visit_sidebar_title (self, node):
        pass

    def depart_sidebar_title (self, node):
        self.cmd (r'\par')
        self.src_sp ()
        self.sp ()

    # Stuff

    def visit_page (self, node):
        self.output_src_sp ()
        self.src_sp ()
        if 'vspace' in node['classes']:
            self.sp (node['length'])
        elif 'clearpage' in node['classes']:
            self.cmd ('\\clearpage\n\n')
        elif 'cleardoublepage' in node['classes']:
            # output a completely blank page
            self.cmd ('\\clearpage{\\pagestyle{empty}\\cleardoublepage}\n\n')
        elif 'vfill' in node['classes']:
            self.cmd ('\\vspace*{\\fill}\n')
        else:
            for matter in ('frontmatter', 'mainmatter', 'backmatter'):
                if matter in node['classes']:
                    self.cmd ('%%\n\\%s\n%%\n\n' % matter)

        raise nodes.SkipNode # empty


    def visit_newline (self, node):
        self.cmd (' \\\\\n')

    def depart_newline (self, node):
        pass


    def visit_attribution (self, node):
        self.cmd (r'\nopagebreak')

    def depart_attribution (self, node):
        self.cmd (r'\par')


    def visit_transition (self, node):
        self.sp ()

    def depart_transition (self, node):
        self.src_sp ()


    def visit_problematic (self, node):
        pass

    def depart_problematic (self, node):
        pass

    def unimplemented_visit (self, node):
        raise NotImplementedError ('visiting unimplemented node type: %s'
                                  % node.__class__.__name__)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

Mydocutils writer package.

Copyright 2010-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""

from __future__ import unicode_literals


__docformat__ = 'reStructuredText'

import collections
import operator

import six
from docutils import nodes, writers
import roman


class Writer (writers.Writer):
    """ A base class for writers. """

    output = None
    """Final translated form of `document`."""

    config_section_dependencies = ('writers', )

    def translate (self):
        visitor = self.translator_class (self.document)
        self.document.walkabout (visitor)
        self.output = visitor.astext ()


class TablePass1 (nodes.SparseNodeVisitor):

    """
    Make a first pass over a table to get a reliable row and column
    count.  Insert placeholder cells for spanned cells.
    """

    def __init__ (self, document):
        nodes.SparseNodeVisitor.__init__ (self, document)

        self.row = -1     # 0-based
        self.column = 0   # 0-based
        self.cells = 0
        self.colspecs = None

    def visit_table (self, table):
        self.colspecs = list (table.traverse (nodes.colspec))
        width = sum (map (operator.itemgetter ('colwidth'), self.colspecs))
        for colspec in self.colspecs:
            colspec['relative_width'] = float (colspec['colwidth']) / width

    def depart_table (self, table):
        table['rows'] = self.rows ()
        table['columns'] = self.cols ()

    def visit_row (self, dummy_node):
        self.row += 1
        self.column = 0
        for colspec in self.colspecs:
            colspec['spanned'] = max (0, colspec.get ('spanned', 0) - 1)

    def visit_entry (self, node):
        """ Table cell. """

        morerows = node.get ('morerows', 0)
        morecols = node.get ('morecols', 0)

        self.cells += (morecols + 1) * (morerows + 1)

        # skip columns that are row-spanned by preceding entries
        while True:
            colspec = self.colspecs [self.column]
            if colspec.get ('spanned', 0) > 0:
                placeholder = nodes.entry ()
                placeholder.type = 'compound'
                placeholder['column'] = self.column
                placeholder.colspecs = self.colspecs[self.column:self.column + 1]
                placeholder['vspan'] = True
                node.replace_self ([placeholder, node])
                self.column += 1
            else:
                break

        # mark columns we row-span
        if morerows:
            for colspec in self.colspecs [self.column : self.column + 1 + morecols]:
                colspec['spanned'] = morerows + 1

        node['row'] = self.row
        node['column'] = self.column

        node.colspecs = self.colspecs[self.column:self.column + morecols + 1]

        self.column += 1 + morecols

        raise nodes.SkipNode

    def rows (self):
        """ Return the no. of columns. """
        return self.row + 1

    def cols (self):
        """ Return the no. of columns. """
        return self.cells // self.rows ()


class ListEnumerator:
    """ Enumerate labels according to list type. """

    def __init__ (self, node, encoding):
        self.type  = node.get ('enumtype') or node.get ('bullet') or '*'
        self.start = node['start'] if 'start' in node else 1
        self.prefix = node.get ('prefix', '')
        self.suffix = node.get ('suffix', '')
        self.encoding = encoding

        self.indent = len (self.prefix + self.suffix) + 1
        if self.type == 'arabic':
            # indentation depends on end value
            self.indent += len (str (self.start + len (node.children)))
        elif self.type.endswith ('alpha'):
            self.indent += 1
        elif self.type.endswith ('roman'):
            self.indent += 5 # FIXME: calculate real length
        else:
            self.indent += 1 # none, bullets, etc.

    def get_next (self):
        """ Get the next label. """

        if self.type == 'none':
            res = ''
        elif self.type == '*':
            res = '•' if self.encoding == 'utf-8' else '-'
        elif self.type == '-':
            res = '-'
        elif self.type == '+':
            res = '+'
        elif self.type == 'arabic':
            res = "%d" % self.start
        elif self.type == 'loweralpha':
            res = "%c" % (self.start + ord ('a') - 1)
        elif self.type == 'upperalpha':
            res = "%c" % (self.start + ord ('A') - 1)
        elif self.type == 'upperroman':
            res = roman.toRoman (self.start).upper ()
        elif self.type == 'lowerroman':
            res = roman.toRoman (self.start).lower ()
        else:
            res = "%d" % self.start

        self.start += 1

        return self.prefix + res + self.suffix

    def get_width (self):
        """ Get indent width for this list. """

        return self.indent


class Translator (nodes.NodeVisitor):
    """ A base translator """

    admonitions = """
    attention caution danger error hint important note tip warning
    """.split ()

    docinfo_elements = """
    address author contact copyright date organization revision status
    version
    """.split ()

    # see http://docutils.sourceforge.net/docs/ref/doctree.html#simple-body-elements

    # simple_structural_subelements = tuple ((getattr (nodes, n) for n in """
    # title subtitle
    # """.split ()))

    # simple_body_elements = tuple ((getattr (nodes, n) for n in """
    # comment doctest_block image literal_block math_block paragraph
    # pending raw rubric substitution_definition target
    # """.split ()))

    # simple_body_subelements = tuple ((getattr (nodes, n) for n in """
    # attribution caption classifier colspec field_name
    # label line option_argument option_string term
    # """.split ()))

    # simple_elements = (simple_structural_subelements +
    #                    simple_body_elements + simple_body_subelements)

    def __init__ (self, document):
        nodes.NodeVisitor.__init__ (self, document)
        self.settings = document.settings

        self.body = []
        self.context = self.body # start with context == body
        self.docinfo = collections.defaultdict (list)
        self.list_enumerator_stack = []
        self.section_level = 0
        self.vspace = 0 # pending space (need this for collapsing)
        self.src_vspace = 0 # pending space for source pretty printing

        self.field_name = None
        self.compacting = 0 # > 0 if we are inside a compacting list
        self.in_literal = 0 # > 0 if we are inside one or more literal blocks

        self.prefixes = collections.defaultdict (list) # dict of arrays of prefixes in order in
                                                       # which to apply classes
        self.suffixes = collections.defaultdict (list) # reverse order of above

        self.environments = [] # stack of \begin'ed environments

        self.register_classes ()

        for name in self.docinfo_elements:
            setattr (self, 'visit_' + name,
                     lambda node: self.visit_field_body (node, name))
            setattr (self, 'depart_' + name, self.depart_field_body)

        for adm in self.admonitions:
            setattr (self, 'visit_' + adm,
                     lambda node: self.visit_admonition (node, adm))
            setattr (self, 'depart_' + adm, self.depart_admonition)


    def register_classes (self):
        pass


    def dispatch_visit (self, node):
        """
        Call self."``visit_`` + node class name" with `node` as
        parameter.  If the ``visit_...`` method does not exist, call
        self.unknown_visit.

        There are 3 hooks for every visit:

        visit_outer
        visit_<classname>
        visit_inner

        """

        self.visit_outer (node)

        node_name = node.__class__.__name__
        method = getattr (self, 'visit_' + node_name, self.unknown_visit)
        self.document.reporter.debug (
            'docutils.nodes.NodeVisitor.dispatch_visit calling %s for %s'
            % (method.__name__, node_name))
        res = method (node)

        if node.type in ('compound', 'simple', 'inline'):
            self.visit_inner (node)

        return res

    def dispatch_departure (self, node):
        """
        Call self."``depart_`` + node class name" with `node` as
        parameter.  If the ``depart_...`` method does not exist, call
        self.unknown_departure.

        There are 3 hooks for every departure:

        depart_inner
        depart_<classname>
        depart_outer

        """

        if node.type in ('compound', 'simple', 'inline'):
            self.depart_inner (node)

        node_name = node.__class__.__name__
        method = getattr (self, 'depart_' + node_name, self.unknown_departure)
        self.document.reporter.debug (
            'docutils.nodes.NodeVisitor.dispatch_departure calling %s for %s'
            % (method.__name__, node_name))
        res = method (node)

        self.depart_outer (node)

        return res


    def unknown_visit (self, node):
        """ Called if we have no handler for this element. """
        pass

    def unknown_departure (self, node):
        """ Called if we have no handler for this element. """
        pass


    def visit_outer (self, node):
        """ The very first hook called on a node, before
        ``visit_<classname>``. """
        pass

    def visit_inner (self, node):
        """ Called after ``visit_<classname>``. """
        pass

    def depart_inner (self, node):
        """ Called on a block before ``depart_<classname>``. """
        pass

    def depart_outer (self, node):
        """ The very last hook called on a node, after
        ``depart_<classname>``."""
        pass


    def register_class (self, types, class_, prefix, suffix):
        """ Register classes.

        A mechanism to automatically output strings before and after
        elements with specific classes.  For most use cases this is
        easier than to write a handler for the element.

        types: types of node this class will apply to:
               tuple of one or more of (text, inline, simple, compound)
        class_: class that triggers the strings
        prefix: string output before element
        suffix: string output after element

        """

        if isinstance (types, six.string_types):
            types = types.split ()

        for t in types:
            self.prefixes[t].append (   (class_, prefix))
            self.suffixes[t].insert (0, (class_, suffix))

    def get_prefix (self, type_, classes):
        return self._get_prefix (type_, classes, self.prefixes)

    def get_suffix (self, type_, classes):
        return self._get_prefix (type_, classes, self.suffixes)

    def _get_prefix (self, type_, classes, array):
        """ Helper for inline handlers. """
        if isinstance (classes, six.string_types):
            classes = classes.split ()

        res = []
        for s in array[type_]:
            if s[0] in classes:
                res.append (s[1])
        return res


    def set_class_on_child (self, node, class_, index = 0):
        """
        Set class `class_` on the visible child no. index of `node`.
        Do nothing if node has fewer children than `index`.
        """
        children = [n for n in node if not isinstance (n, nodes.Invisible)]
        try:
            child = children[index]
        except IndexError:
            return
        child['classes'].append (class_)

    def set_first_last (self, node):
        """ Set class 'first' on first child, 'last' on last child. """
        self.set_class_on_child (node, 'first', 0)
        self.set_class_on_child (node, 'last', -1)

    def astext (self):
        """ Return the final formatted document as a string. """
        return self.preamble () + ''.join (self.context) + self.postamble ()

    def comment (self, text):
        """ Output a comment. """
        pass

    def text (self, text):
        """ Output text. """
        pass

    def sp (self, n = 1):
        """ Adds vertical space before the next simple element.

        All spaces added collapse into the largest one. """

        if n == 0:
            self.vspace = 1999
        else:
            self.vspace = max (n, self.vspace)

    def src_sp (self, n = 1):
        """ Add vertical space to the source. """

        if n == 0:
            self.src_vspace = 1999
        else:
            self.src_vspace = max (n, self.src_vspace)

    def output_sp (self):
        pass

    def output_src_sp (self):
        pass

    def push (self):
        """ Push environment. """
        pass

    def pop (self):
        """ Pop environment. """
        pass

    def br_if_line_longer_than (self, length):
        """ Go one line up if the last line was shorter than length.

        Use this to compact lists etc. """
        pass

    def indent (self, by = 2):
        """ Indent text. """
        pass

    def rindent (self, by = 2):
        """ Indent text on the right side. """
        pass

    def preamble (self):
        return ''

    def postamble (self):
        return ''

    def visit_title (self, node):
        """ Switch on the various incarnations the title element can have. """

        if isinstance (node.parent, nodes.section):
            self.visit_section_title (node)
        elif isinstance (node.parent, nodes.document):
            self.visit_document_title (node)
        elif isinstance (node.parent, nodes.table):
            self.visit_table_title (node)
        elif isinstance (node.parent, nodes.topic):
            self.visit_topic_title (node)
        elif isinstance (node.parent, nodes.sidebar):
            self.visit_sidebar_title (node)
        elif isinstance (node.parent, nodes.admonition):
            self.visit_admonition_title (node)
        else:
            assert ("Can't happen.")

    def depart_title (self, node):
        """ Switch on the various incarnations the title element can have. """

        if isinstance (node.parent, nodes.section):
            self.depart_section_title (node)
        elif isinstance (node.parent, nodes.document):
            self.depart_document_title (node)
        elif isinstance (node.parent, nodes.table):
            self.depart_table_title (node)
        elif isinstance (node.parent, nodes.topic):
            self.depart_topic_title (node)
        elif isinstance (node.parent, nodes.sidebar):
            self.depart_sidebar_title (node)
        elif isinstance (node.parent, nodes.admonition):
            self.depart_admonition_title (node)
        else:
            assert ("Can't happen.")

    def visit_subtitle (self, node):
        """ Switch on the various incarnations the subtitle element can have. """

        if isinstance (node.parent, nodes.document):
            self.visit_document_subtitle (node)
        else:
            self.visit_section_subtitle (node)

    def depart_subtitle (self, node):
        """ Switch on the various incarnations the subtitle element can have. """

        if isinstance (node.parent, nodes.document):
            self.depart_document_subtitle (node)
        else:
            self.depart_section_subtitle (node)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
TxtPackager.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Package a Txt file for PG.

"""

from ebookmaker.packagers import OneFileZipPackager

TYPE = 'ww'
FORMATS = 'txt.us-ascii txt.iso-8859-1 txt.utf-8'.split ()

class Packager (OneFileZipPackager):
    """ WW packager for plain text files. """
    pass



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
TxtWriter.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Build an UTF-8-encoded PG plain text file. This is just the plain text
version recoded into UTF-8.

"""

from __future__ import unicode_literals

import os
import subprocess
import sys

from libgutenberg.Logger import debug, info, warning, error
from libgutenberg.GutenbergGlobals import SkipOutputFormat, mkdir_for_filename

from ebookmaker import ParserFactory
from ebookmaker import writers
from ebookmaker.CommonCode import Options

options = Options()

# map some not-widely-supported characters to more common ones
u2u = {
    0x2010: '-',  # unicode HYPHEN to HYPHEN-MINUS. Many Windows fonts lack this.
    }

class Writer(writers.BaseWriter):
    """ Class to write PG plain text. """

    def groff(self, job, nroff, encoding='utf-8'):
        """ Process thru groff.

        Takes and returns unicode strings!

        """

        device = {'utf-8': 'utf8',
                  'iso-8859-1': 'latin1',
                  'us-ascii': 'ascii'}[encoding]

        nroff = nroff.encode(encoding)
        nrofffilename = os.path.join(
            os.path.abspath(job.outputdir),
            os.path.splitext(job.outputfile)[0] + '.nroff')

        # write nroff file for debugging
        if options.verbose >= 2:
            with open(nrofffilename, 'wb') as fp:
                fp.write(nroff)
        else:
            try:
                # remove debug files from previous runs
                os.remove(nrofffilename)
            except OSError:
                pass

        # call groff
        try:
            _groff = subprocess.Popen([options.config.GROFF,
                                       "-t",             # preprocess with tbl
                                       "-K", device,     # input encoding
                                       "-T", device],    # output device
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE)
        except OSError:
            error("TxtWriter: executable not found: %s" % options.config.GROFF)
            raise SkipOutputFormat

        (txt, stderr) = _groff.communicate(nroff)

        # pylint: disable=E1103
        for line in stderr.splitlines():
            line = line.decode(sys.stderr.encoding)
            line = line.strip()
            if 'error' in line:
                error("groff: %s" % line)
            elif 'warn' in line:
                if options.verbose >= 1:
                    warning("groff: %s" % line)

        txt = txt.decode(encoding)
        return txt.translate(u2u) # fix nroff idiosyncracies


    def build(self, job):
        """ Build TXT file. """

        filename = os.path.join(job.outputdir, job.outputfile)

        encoding = job.subtype.strip('.')

        mkdir_for_filename(filename)

        info("Creating plain text file: %s from %s", filename, job.url)

        parser = ParserFactory.ParserFactory.create(job.url)

        # don't make txt file unless the source is txt of some encoding
        has_txt_source = 'text/plain' in str(parser.attribs.orig_mediatype)
        is_html_source = not has_txt_source and \
                         hasattr(parser, 'xhtml') and \
                         parser.xhtml is not None

        if hasattr(parser, 'rst2nroff'):
            data = self.groff(job, parser.rst2nroff(job, encoding), encoding)
        elif is_html_source:
            info("Plain text file %s aborted due to html input" % filename)
            return
        else:
            data = parser.unicode_content()

        data = data.encode('utf_8_sig' if encoding == 'utf-8' else encoding, 'unitame')

        self.write_with_crlf(filename, data)

        info("Done plain text file: %s" % filename)
